/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.basis;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import de.bernhardjung.dsaprogramm.basiswerte.Basiswert;
import de.bernhardjung.dsaprogramm.gui.DialogFehlermeldung;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.Sonderfertigkeit;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.SonderfertigkeitWert;
import de.bernhardjung.dsaprogramm.talente.Kampftechnik;
import de.bernhardjung.dsaprogramm.util.Util;

public class FernkampfwaffeWert implements Cloneable {
    
    /**
     * Erzeugt eine Deep Copy der Fernkampfwaffen.
     */
    public static ArrayList< FernkampfwaffeWert > deepCopy(
            final Collection< FernkampfwaffeWert > originale ) {
        
        final ArrayList< FernkampfwaffeWert > kopien =
            new ArrayList< FernkampfwaffeWert >( originale.size() );
        
        for ( final FernkampfwaffeWert waffe : originale ) {
         
            kopien.add(  waffe.clone() );
        }
        
        return kopien;
    }

    /** Der Typ der Fernwaffe, um den es sich handelt. */
    private Fernkampfwaffe fernkampfwaffe;

    /** Die Bezeichnung der Waffe. */
    private String name;

    /**
     * Die Reichweiten der Waffe. Sehr nah steht an Position 0, extrem weit an
     * Position 4.
     */
    private List< Integer > reichweiten;

    /** Das Talent mit dem die Waffe gefuehrt wird. */
    private Kampftechnik talent;

    /** Die Trefferpunkte der Waffe. */
    private Tp tp;

    /**
     * Die Trefferpunkte durch Entfernung. Die TP fuer sehr nah stehen an
     * Position 0, die fuer extrem weit an Position 4.
     */
    private List< Integer > tpPlus;

    public FernkampfwaffeWert() {
        this( new Fernkampfwaffe() );
    }

    public FernkampfwaffeWert( final Fernkampfwaffe dieFernwaffe ) {

        reichweiten = new ArrayList< Integer >( 5 );

        tpPlus = new ArrayList< Integer >( 5 );

        setWaffe( dieFernwaffe );
    }

    /**
     * Erzeugt eine Deep Copy des Wertes.
     */
    @Override
    protected FernkampfwaffeWert clone() {
        
        FernkampfwaffeWert clone = null;
        
        try {
            
            clone = (FernkampfwaffeWert) super.clone();
         
            clone.reichweiten = new ArrayList< Integer >( reichweiten );
            
            clone.tpPlus =  new ArrayList< Integer >( tpPlus );
            
        } catch ( CloneNotSupportedException e ) {
            
            DialogFehlermeldung.zeigeDialog( e );
        }
        
        return clone;
    }
    
    @Override
    public boolean equals( Object obj ) {
        
        if ( null == obj 
                || ! obj.getClass().equals( getClass() ) )
        {
            return false;
        }
        
        final FernkampfwaffeWert wert = (FernkampfwaffeWert) obj;
        
        return Util.equals( 
                fernkampfwaffe, wert.fernkampfwaffe,
                name, wert.name,
                reichweiten, wert.reichweiten,
                talent, wert.talent,
                tp, wert.tp,
                tpPlus, wert.tpPlus );
    }
    
    /**
     * Liefert den Fernkampfwert eines Helden mit dieser Waffe.
     *
     * @param held
     *            dessen Fernkampfwert gesucht ist.
     * @return den Fernkampfwert des Helden.
     */
    public final int getFernkampfwert( final Held held ) {

        int fernkampfwert = held.getStufe( Basiswert.FKB );

        if ( held.hat( talent ) ) {

            fernkampfwert += held.getStufe( talent );
        }

        if ( ! held.hat( talent )
          || ! fernkampfwaffe.hatHaupttalent( talent ) ) {

            fernkampfwert -= 5;
        }

        final Sonderfertigkeit spezialisierung =
            Sonderfertigkeit.getWaffenspezialisierung( talent );

        final SonderfertigkeitWert wert = new SonderfertigkeitWert( 
                spezialisierung, 
                fernkampfwaffe.getLokalisierungsschluesselName() );

        if ( held.hat( wert ) ) {

            fernkampfwert += 2;
        }

        int be = held.getBehinderung();

        be = be - talent.getEBe();

        if ( 0 < be ) {

            fernkampfwert -= be;
        }

        return fernkampfwert;
    }

    /**
     * @return der Typ der Fernwaffe, um den es sich hier handelt.
     */
    public final Fernkampfwaffe getWaffe() {
        return fernkampfwaffe;
    }

    /**
     * @return den Namen der Waffe.
     */
    public String getName() {
        return name;
    }

    /**
     * @return die Reichweiten der Waffe. Sehr nah ist an Position 0, extrem
     *         weit an Position 4.
     */
    public List< Integer > getReichweiten() {
        return new ArrayList< Integer >( reichweiten );
    }

    /** @return das Talent mit dem die Waffe gefuehrt wird. */
    public Kampftechnik getTalent() {
        return talent;
    }

    /** @return die TP der Waffe. */
    public Tp getTp() {
        return tp;
    }

    /**
     * @return die TP+ der Waffe. Sehr nah ist an Position 0, extrem
     *         weit an Position 4.
     */
    public List< Integer > getTpPlus() {
        return new ArrayList< Integer >( tpPlus );
    }

    /**
     * Bestimmt den Typ der Fernwaffe, um die es sich handelt. Die Eigenschaften
     * der Fernwaffe werden uebernommen. Also Gewicht, Reichweiten usw. werden
     * auf die Werte der Fernwaffe eingestellt.
     *
     * @param dieFernwaffe
     *            um die es sich bei dieser Waffe handelt.
     */
    public final void setWaffe( final Fernkampfwaffe dieFernwaffe ) {

        if ( null == fernkampfwaffe ) {
            new IllegalArgumentException(
                    "Es wurde null fuer dieFernwaffe uebergeben." );
        }

        fernkampfwaffe = dieFernwaffe;

        setName( fernkampfwaffe.getName() );

        setReichweiten( fernkampfwaffe.getReichweiten() );

        setTalent( fernkampfwaffe.getTalent() );

        setTp( fernkampfwaffe.getTp() );

        setTpPlus( fernkampfwaffe.getTpPlus() );
    }

    /**
     * @param neuerName der Waffe.
     */
    public void setName( final String neuerName ) {
        this.name = neuerName;
    }

    /**
     * @param neueReichweiten fuer die Waffe. Sehr nah steht an Position 0,
     *          extrem weit an Position 4.
     */
    public void setReichweiten( final List< Integer > neueReichweiten ) {
        reichweiten.clear();

        reichweiten.addAll( neueReichweiten );
    }

    /**
     * @param technik Das Talent mit dem die Waffe gefuehrt werden soll.
     */
    public void setTalent( final Kampftechnik technik ) {
        talent = technik;
    }

    /**
     * @param neueTp der Waffe.
     */
    public void setTp( final Tp neueTp ) {
        tp = neueTp;
    }

    /**
     * @param neueTpPlus fuer die Waffe. Sehr nah steht an Position 0,
     *          extrem weit an Position 4.
     */
    public void setTpPlus( final List< Integer > neueTpPlus ) {
        tpPlus.clear();

        tpPlus.addAll( neueTpPlus );
    }

    @Override
    public String toString() {
        return name;
    }
}
