/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.basis.heldio;

import java.io.File;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.List;

import com.lowagie.text.BadElementException;
import com.lowagie.text.Cell;
import com.lowagie.text.Document;
import com.lowagie.text.DocumentException;
import com.lowagie.text.Element;
import com.lowagie.text.Font;
import com.lowagie.text.Paragraph;
import com.lowagie.text.Table;
import com.lowagie.text.rtf.RtfWriter2;
import com.lowagie.text.rtf.style.RtfFont;

import de.bernhardjung.dsaprogramm.basis.FernkampfwaffeWert;
import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.basis.NahkampfwaffeWert;
import de.bernhardjung.dsaprogramm.basis.RuestungWert;
import de.bernhardjung.dsaprogramm.basis.SKT;
import de.bernhardjung.dsaprogramm.basis.SchildWert;
import de.bernhardjung.dsaprogramm.basis.Tp;
import de.bernhardjung.dsaprogramm.basiswerte.Basiswert;
import de.bernhardjung.dsaprogramm.basiswerte.BasiswertWert;
import de.bernhardjung.dsaprogramm.eigenschaften.Eigenschaft;
import de.bernhardjung.dsaprogramm.eigenschaften.EigenschaftWert;
import de.bernhardjung.dsaprogramm.gui.DialogFehlermeldung;
import de.bernhardjung.dsaprogramm.i18n.I18n;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.Sonderfertigkeit;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.SonderfertigkeitWert;
import de.bernhardjung.dsaprogramm.talente.Kampftechnik;
import de.bernhardjung.dsaprogramm.talente.Talent;
import de.bernhardjung.dsaprogramm.talente.TalentWert;
import de.bernhardjung.dsaprogramm.talente.Talentgruppe;
import de.bernhardjung.dsaprogramm.util.Strings;
import de.bernhardjung.dsaprogramm.vornachteile.VorNachteil;
import de.bernhardjung.dsaprogramm.vornachteile.VorNachteilWert;
import de.bernhardjung.dsaprogramm.zauber.Zauber;
import de.bernhardjung.dsaprogramm.zauber.ZauberWert;
import de.bernhardjung.dsaprogramm.zauber.Zaubergruppe;

public class HeldExportRtf {

    /** Der Text in der Zeile in der das AP-Guthaben des Helden steht. */
    private static final String AP_GUTHABEN = "HeldExportRtf.ApGuthaben";

    /**
     * Die Ueberschrift in der Talente Tabelle fuer die Spalte in der die
     * Punkte stehen die auf Attacke verteilt wurden.
     */
    private static final String ATTACKE = "HeldExportRtf.Talente.AT";

    /**
     * Die Ueberschrift in der Waffentabelle fuer die Spalte in der der
     * Attackewert steht.
     */
    private static final String AT = "HeldExportRtf.Kampf.At";

    /** Der Text der neben der Augenfarbe des Helden steht. */
    private static final String AUGENFARBE = "HeldExportRtf.Augenfarbe";

    /** Der Text der neben der Beschreibung des Aussehens des Heldens steht. */
    private static final String AUSSEHEN = "HeldExportRtf.Aussehen";

    /**
     * Die Ueberschrift fuer den Abschnitt in dem der Ausweichenwert steht.
     */
    private static final String AUSWEICHEN = "HeldExportRtf.Kampf.Ausweichen";

    /**
     * Die Ueberschrift in der Basiswerter Tabelle fuer die Spalte in der die
     * aktuellen Stufen stehen.
     */
    private static final String BASISWERT_AKTUELL = "HeldExportRtf.Basiswert.Aktuell";

    /**
     * Die Ueberschrift in der Basiswerter Tabelle fuer die Spalte in der die
     * maximal zukaufbaren Punkte stehen.
     */
    private static final String BASISWERT_MAX_ZUGEKAUFT = "HeldExportRtf.Basiswert.MaxZugekauft";

    /**
     * Die Ueberschrift in der Basiswerter Tabelle fuer die Spalte in der die
     * Modifikation steht.
     */
    private static final String BASISWERT_MOD = "HeldExportRtf.Basiswert.Mod";

    /**
     * Die Ueberschrift in der Basiswerter Tabelle fuer die Spalte in welcher
     * die Basisstufen stehen. Das sind die Werte die sich nur aus der
     * Berechnung der Formel ergeben.
     */
    private static final String BASISWERT_BASISSTUFE = "HeldExportRtf.Basiswert.Basisstufe";

    /**
     * Die Ueberschrift in der Basiswerter Tabelle fuer die Spalte in welcher
     * die zugekauften Punkte stehen.
     */
    private static final String BASISWERT_ZUGEKAUFT = "HeldExportRtf.Basiswert.Zugekauft";

    /**
     * Die Ueberschrift in der Ruestungen-Tabelle fuer die Spalte in der die
     * Behinderung steht.
     */
    private static final String BE = "HeldExportRtf.Kampf.Be";

    /**
     * Die Ueberschrift fuer den Abschnitt in dem der Bruchfaktor der Waffe/
     * des Schildes steht.
     */
    private static final String BRUCHFAKTOR = "HeldExportRtf.Kampf.Bruchfaktor";

    /**
     * Die Ueberschrift fuer die Spalte mit der Distanzklasse in der
     * Nahkampfwaffentabelle.
     */
    private static final String DK = "HeldExportRtf.Kampf.Dk";

    /**
     * Die Ueberschrift in der Eigenschaften Tabelle fuer die Spalte in welcher
     * die aktuelle Stufe steht.
     */
    private static final String EIGENSCHAFT_AKTUELL = "HeldExportRtf.Eigenschaft.Aktuell";

    /**
     * Die Ueberschrift in der Eigenschaften Tabelle fuer die Spalte in welcher
     * die Startstufe steht.
     */
    private static final String EIGENSCHAFT_STARTWERT = "HeldExportRtf.Eigenschaft.Startwert";

    /**
     * Die Ueberschrift in der Eigenschaften Tabelle fuer die Spalte in welcher
     * die zugekauften Punkte stehen.
     */
    private static final String EIGENSCHAFT_ZUKAUF = "HeldExportRtf.Eigenschaft.Zukauf";

    /**
     * Die Ueberschrift fuer die Tabellen mit den Eigenschaften und den
     * Basiswerten.
     */
    private static final String EIGENSCHAFTEN_BASISWERTE = "HeldExportRtf.EigenschaftenBasiswerte";

    /** Der Text in der Zeile in der die eingesetzten AP des Helden stehen. */
    private static final String EINGESETZTE_AP = "HeldExportRtf.EingesetzteAp";

    /**
     * Die Ueberschrift fuer die Spalte mit der Entfernung in der
     * Fernkampfwaffentabelle.
     */
    private static final String ENTFERNUNGEN = "HeldExportRtf.Kampf.Entfernungen";

    /**
     * Die Ueberschrift fuer die Spalte mit dem Namen der Fernkampfwaffe.
     */
    private static final String FERNKAMPFWAFFE = "HeldExportRtf.Kampf.Fernkampfwaffe";

    /**
     * Die Ueberschrift fuer die Spalte mit dem Fernkampfwert in der
     * Fernkampfwaffentabelle.
     */
    private static final String FK = "HeldExportRtf.Kampf.Fk";

    /**
     * Der Text der neben dem Geburtstag des Helden steht.
     */
    private static final String GEBURTSTAG = "HeldExportRtf.Geburtstag";

    /** Der Text neben den Gesamt-AP des Helden. */
    private static final String GESAMT_AP = "HeldExportRtf.GesamtAp";

    /**
     * Der Text der neben dem Geschlecht des Helden steht.
     */
    private static final String GESCHLECHT = "HeldExportRtf.Geschlecht";

    /**
     * Die Ueberschrift fuer die Spalte mit der Anzahl der Geschosse in der
     * Nahkampfwaffentabelle.
     */
    private static final String GESCHOSSE = "HeldExportRtf.Kampf.Geschosse";

    /**
     * Das Wort 'Geschwindigkeit' zum Angeben der Geschwindigkeit.
     */
    private static final String GESCHWINDIGKEIT = "HeldExportRtf.Geschwindigkeit";

    /**
     * Der Text der neben dem Gewicht des Helden steht.
     */
    private static final String GEWICHT = "HeldExportRtf.Gewicht";

    /** Der Text der neben der Groesse des Helden steht. */
    private static final String GROESSE = "HeldExportRtf.Groesse";

    /** Der Text der neben der Haarfarbe des Helden steht. */
    private static final String HAARFARBE = "HeldExportRtf.Haarfarbe";

    /** Die Einheit fuer die Koerpergroesse. */
    private static final String HALBFINGER = "HeldExportRtf.Halbfinger";

    /** Die Ueberschrift des Dokuments. */
    private static final String HELDENDOKUMENT = "HeldExportRtf.Heldendokument";

    /**
     * Der Text der die Beschreibung von Familie, Herkunft und Hintergrund des
     * Helden markiert..
     */
    private static final String HINTERGRUND = "HeldExportRtf.Hintergrund";

    /**
     * Die Ueberschrift fuer die Spalte mit der Initiative in der
     * Nahkampfwaffentabelle.
     */
    private static final String INI = "HeldExportRtf.Kampf.Ini";

    /**
     * Die Ueberschrift fuer den Abschnitt fuer die Initiative.
     */
    private static final String INITIATIVE = "HeldExportRtf.Kampf.Initiative";

    /** Der Text der neben der Kultur des Helden steht. */
    private static final String KULTUR = "HeldExportRtf.Kultur";

    /**
     * Die Ueberschrift fuer die Spalte mit dem Namen der Waffe in der
     * Nahkampfwaffentabelle.
     */
    private static final String NAHKAMPFWAFFE = "HeldExportRtf.Kampf.Nahkampfwaffe";

    /** Der Text der neben dem Namen des Helden steht. */
    private static final String NAME = "HeldExportRtf.Name";

    /**
     * Die Uebeschrift in der Talente Tabelle fuer die Spalte in der die
     * Punkte stehen, die auf Parade verteilt wurden.
     */
    private static final String PARADE = "HeldExportRtf.Talente.PA";

    /**
     * Die Ueberschrift in der Waffentabelle fuer die Spalte in der der
     * Paradewert steht.
     */
    private static final String PA = "HeldExportRtf.Kampf.Pa";

    /** Der Text der neben der Profession des Helden steht. */
    private static final String PROFESSION = "HeldExportRtf.Profession";

    /** Der Text der die Rasse des Helden markiert. */
    private static final String RASSE = "HeldExportRtf.Rasse";

    /**
     * Die Ueberschrift in der Ruestungentabelle fuer die Spalte in der der
     * Ruestungsschutz steht.
     */
    private static final String RS = "HeldExportRtf.Kampf.Rs";

    /**
     * Die Ueberschrift fuer die Ruestungstabelle.
     */
    private static final String RUESTUNG = "HeldExportRtf.Kampf.Ruestung";

    /**
     * Die Ueberschrift in der Ruestungstabelle fuer die Spalte in der der
     * Name der Ruestung steht.
     */
    private static final String RUESTUNGSSTUECK = "HeldExportRtf.Kampf.Ruestungsstueck";

    /**
     * Die Ueberschrift fuer die Schildtabelle.
     */
    private static final String SCHILD_PARIERWAFFE = "HeldExportRtf.Kampf.SchildParierwaffe";

    /**
     * Die Ueberschrift fuer die Spalte mit den Sonderfertigkeiten in der
     * Ausweichen Tabelle.
     */
    private static final String SONDERFERTIGKEIT = "HeldExportRtf.Kampf.Sonderfertigkeit";

    /** Die Ueberschrift fuer die Aufzaehlung der Sonderfertigkeiten. */
    private static final String SONDERFERTIGKEITEN = "HeldExportRtf.Sonderfertigkeiten";

    /** Der Text der neben der Beschreibung des Standes des Helden steht. */
    private static final String STAND = "HeldExportRtf.Stand";

    /** Die Einheit fuer das Gewicht des Helden. */
    private static final String STEIN = "HeldExportRtf.Stein";

    /** Fuer die SUmme der BE durch Ruestungen. */
    private static final String SUMME = "HeldExportRtf.Kampf.Summe";

    /** Die Uebeschrift fuer die Tabelle mit den Talenten des Helden. */
    private static final String TALENTE = "HeldExportRtf.Talente";

    /**
     * Die Uebeschrift in der Tabelle mit den Talenten fuer die Spalte in der
     * die Stufe des Talents steht.
     */
    private static final String TAW = "HeldExportRtf.Talente.TaW";

    /** Der Text der neben dem Titel des Helden steht. */
    private static final String TITEL = "HeldExportRtf.Titel";

    /**
     * Die Ueberschrift fuer Spalte mit den TP in den Waffentabellen.
     */
    private static final String TP = "HeldExportRtf.Kampf.Tp";

    /**
     * Die Ueberschrift fuer Spalte mit den TP (A) in der Nahkampftabelle.
     */
    private static final String TP_A = "HeldExportRtf.Kampf.TpA";

    /**
     * Die Ueberschrift fuer Spalte mit den TP durch Entfernung in der
     * Fernkampftabelle.
     */
    private static final String TP_ENTFERNUNG = "HeldExportRtf.Kampf.TpEntfernung";

    /**
     * Die Ueberschrift fuer Spalte mit den TP+ in der Nahkampfwaffentabelle.
     */
    private static final String TP_KK = "HeldExportRtf.Kampf.TpKk";

    /**
     * Die Ueberschrift fuer Spalte mit dem Typ in der Schildetabelle.
     */
    private static final String TYP = "HeldExportRtf.Kampf.Typ";

    /**
     * Die Ueberschrift fuer Spalte mit den effektiven Behinderung in der
     * Nahkampfwaffentabelle.
     */
    private static final String TYP_EBE = "HeldExportRtf.Kampf.TypEbe";

    /** Die Ueberschrift für die Aufzaehlung der Vor-/Nachteile. */
    private static final String VOR_NACHTEILE = "HeldExportRtf.VorNachteile";

    /**
     * Die Ueberschrift fuer die Waffentabellen.
     */
    private static final String WAFFEN_KAMPFWERTE = "HeldExportRtf.Kampf.WaffenKampfwerte";

    /**
     * Die Ueberschrift fuer die Spalten mit dem Namen der Kampftechnik in der
     * Nahkampftabellen.
     */
    private static final String WAFFENLOSER_KAMPF = "HeldExportRtf.Kampf.WaffenloserKampf";

    /**
     * Die Ueberschrift fuer die Spalten mit dem Waffenmodifikator in der
     * Nahkampfwaffentabellen.
     */
    private static final String WM = "HeldExportRtf.Kampf.Wm";

    /**
     * Die Uebeschrift, in der Tabelle mit den Zaubern, fuer die Spalte in der
     * die Stufe eines Zaubers steht.
     */
    private static final String ZFW = "HeldExportRtf.Talente.ZfW";

    /**
     * Die Anzahl Zeilen der Talente-Tabelle die auf eine Seite passen.
     */
    private static final int ZEILEN_PRO_SEITE = 35;

    private final Font schrift22b = new RtfFont( "Times", 22, RtfFont.BOLD );
    private final Font schrift16b = new RtfFont( "Times", 16, RtfFont.BOLD );
    private final Font schrift12b = new RtfFont( "Times", 12, RtfFont.BOLD );
    private final Font schrift11b = new RtfFont( "Times", 11, RtfFont.BOLD );
    private final Font schrift10b = new RtfFont( "Times", 10, RtfFont.BOLD );
    private final Font schrift6b  = new RtfFont( "Times", 6,  RtfFont.BOLD );

    private final Font schrift10 = new RtfFont( "Times", 10 );
    private final Font schrift12 = new RtfFont( "Times", 12 );

    /**
     * Fuegt die Kampftechniken in die Tabelle mit den Talentwerten ein.
     * Es wird davon ausgegangen das die Kampftechniken ganz am Anfang der
     * Tabelle stehen.
     *
     * @param tabelle in die Kampftechniken eingetragen werden sollen.
     * @param held dessen Werte in der Tabelle dargestellt werden.
     * @return die naechste freie Zeile in der Tabelle nach den Kampftechniken.
     * @throws BadElementException
     */
    private int einfuegenKampftechniken(
            final Table tabelle,
            final Held held ) throws BadElementException {

        final Talentgruppe gruppe = Talentgruppe.KAMPF;

        int zeile  = 0;

        final int spalte = 0;

        Cell zelle;
        zelle = erzeugeZelle( gruppe.getName(), schrift12b, Element.ALIGN_LEFT);

        zelle.setColspan( 2 );

        tabelle.addCell( zelle, zeile, spalte );

        zelle = erzeugeZelle(
                ATTACKE,
                schrift12b,
                Element.ALIGN_LEFT );

        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                PARADE,
                schrift12b,
                Element.ALIGN_LEFT );

        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                TAW,
                schrift12b,
                Element.ALIGN_CENTER );

        zelle.setColspan( 2 );

        tabelle.addCell( zelle );

        List< Talent > kampftechniken = held.getTalente( gruppe );
        kampftechniken.sort( Talent.NAME_COMPARATOR );
        for ( final Talent talent : kampftechniken ) {

            ++zeile;

            if ( held.hat( VorNachteil.ELFISCHE_WELTSICHT )
              && held.hatLeittalent( talent ) ) {

                zelle = erzeugeZelle(
                        talent.getName() + "*",
                        schrift10,
                        Element.ALIGN_LEFT );

            } else {

                zelle = erzeugeZelle(
                        talent.getName(),
                        schrift10,
                        Element.ALIGN_LEFT );

            }

            setRahmen( zelle, 0.1F, 0F, 0.1F, 0.1F );

            tabelle.addCell( zelle, zeile, spalte );

            final int stufe = held.getStufe( talent );

            final int faktor =
                talent.getSteigerungsspalte( held, stufe );

            zelle = erzeugeZelle(
                    SKT.toString( faktor ),
                    schrift10,
                    Element.ALIGN_RIGHT );

            setRahmen( zelle, 0.1F, 0.1F, 0.1F, 0F );

            tabelle.addCell( zelle );

            final TalentWert wert = held.getWert( talent );

            zelle = erzeugeZelle(
                    String.valueOf( wert.getAttacke() ),
                    schrift10,
                    Element.ALIGN_RIGHT );

            tabelle.addCell( zelle );

            zelle = erzeugeZelle(
                    String.valueOf( wert.getParade( held ) ),
                    schrift10,
                    Element.ALIGN_RIGHT );

            tabelle.addCell( zelle );

            final String stufeString;

            if ( stufe < 0 && ( ! talent.istBasisTalent() ) ) {

                stufeString = String.format( "(%1$d)", stufe );

            } else {

                stufeString = String.valueOf( stufe );

            }

            zelle = erzeugeZelle(
                    stufeString,
                    schrift10,
                    Element.ALIGN_RIGHT );

            tabelle.addCell( zelle );

            zelle = erzeugeZelle( " ", schrift10, Element.ALIGN_RIGHT );

            tabelle.addCell( zelle );
        }

        ++zeile;
        // zeile zeigt jetzt auf die erste Zeile hinter den Kampftechniken.

        return zeile;
    }

    /**
     * Fuegt in eine Tabelle Teiltabellen mit Talenten ein.
     * Die Tabelle muss 13 Spalten haben. Die linken 6 und die rechten 6
     * Spalten werden jeweils fuer eine Teiltabelle benutzt.
     *
     * @param tabelle die Haupttabelle in die die Teiltabellen eingefuegt
     *      werden.
     * @param held dessen Werte in der Tabelle dargstellt werden.
     * @param talente die in der Tabelle dargestellt werden.
     * @param zeile der Haupttabelle ab der die Teiltabellen eingefuegt werden.
     *      Die erste Zeile ist Zeile 0.
     * @throws BadElementException
     */
    private Tabellenende einfuegenTalente(
            final Table tabelle,
            final Held held,
            final TalenteTabelleDaten talente,
            int zeile ) throws BadElementException {

        int spalte = 0;

        while ( talente.hatTalentgruppe() ) {

            final int freieZeilen = ZEILEN_PRO_SEITE - zeile;

            if ( 2 > freieZeilen ) {
                // Es sind nicht mehr genug freie Zeilen da um eine neue
                // Teiltabelle zu beginnen.
                if ( 0 == spalte ) {
                    // Die linke Haelfte der Haupttabelle ist gefuellt.
                    // Es kann jetzt die rechte Haelfte gefuellt werden.
                    spalte = 7;
                    zeile = 0;

                } else {

                    // Die ganze Haupttabelle ist voll.
                    // Es kann abgebrochen werden.
                    return new Tabellenende( zeile, spalte );

                }
            }

            final Talentgruppe talentgruppe = talente.getTalentgruppe();

            Cell zelle;

            zelle = erzeugeZelle(
                    talentgruppe.getName(),
                    schrift12b,
                    Element.ALIGN_LEFT );

            zelle.setColspan( 4 );

            tabelle.addCell( zelle, zeile, spalte );

            zelle = erzeugeZelle(
                    TAW,
                    schrift12b,
                    Element.ALIGN_CENTER );

            zelle.setColspan( 2 );

            tabelle.addCell( zelle );

            ++zeile;

            while ( talente.hatNaechstesTalent() ) {

                final Talent talent = talente.getNaechstesTalent();

                final StringBuilder text = new StringBuilder();

                text.append( talent.getName() );


                if ( held.hat( VorNachteil.ELFISCHE_WELTSICHT )
                  && held.hatLeittalent( talent ) ) {

                    text.append( "* " );

                } else {

                    text.append( " " );

                }

                text.append( talent.getProbe().toString() );

                zelle = erzeugeZelle(
                        text.toString(),
                        schrift10,
                        Element.ALIGN_LEFT );

                zelle.setColspan( 3 );

                setRahmen( zelle, 0.1F, 0F, 0.1F, 0.1F );

                tabelle.addCell( zelle, zeile, spalte );

                final int stufe = held.getStufe( talent );

                final int faktor =
                    talent.getSteigerungsspalte( held, stufe );

                zelle = erzeugeZelle( SKT.toString( faktor ),
                                      schrift10,
                                      Element.ALIGN_RIGHT );

                setRahmen( zelle, 0.1F, 0.1F, 0.1F, 0F );

                tabelle.addCell( zelle );

                final String stufeString;

                if ( stufe < 0 && ( ! talent.istBasisTalent() ) ) {

                    stufeString = String.format( "(%1$d)", stufe );

                } else {

                    stufeString = String.valueOf( stufe );

                }

                zelle = erzeugeZelle( stufeString,
                                      schrift10,
                                      Element.ALIGN_RIGHT );

                tabelle.addCell( zelle );

                zelle = erzeugeZelle( " ", schrift10, Element.ALIGN_RIGHT );

                tabelle.addCell( zelle );

                ++zeile;

                if ( ZEILEN_PRO_SEITE == zeile ) {
                    // Eine Haelfte der Tabelle ist komplett gefuellt
                    // vielleicht auch beide.

                    if ( 0 == spalte ) {
                        // Die linke Haelfte der Haupttabelle ist gefuellt, jetzt
                        // wird die rechte Seite gefuellt.
                        spalte = 7;
                        zeile = 0;
                        break;

                    } else {

                        if ( ! talente.hatNaechstesTalent() ) {
                            talente.naechsteTalentgruppe();
                        }
                        // Die Tabelle ist komplett gefuellt.
                        // Es kann abgebrochen werden
                        return new Tabellenende( zeile, spalte );
                    }
                }
            }

            if ( ! talente.hatNaechstesTalent() ) {
                // Die innere Schleife hat alle Talente der Talentgruppe
                // abgearbeitet.
                talente.naechsteTalentgruppe();

                if ( 0 != zeile ) {
                    // Es ist nicht der Anfang der Haupttabelle.
                    // Es kommt eine Teiltabelle vorher.
                    // Zu dieser wird ein Leerzeile Platz gelassen.
                    ++zeile;
                }
            }
        }

        return new Tabellenende( zeile, spalte );
    }

    /**
     * Fuegt in eine Tabelle eine Teiltabelle mit Zaubern. Die Tabelle muss 13
     * Spalten haben. Die linken 6 und die rechten 6 Spalten werden jeweils fuer
     * eine Teiltabelle benutzt.
     *
     * @param tabelle
     *            die Haupttabelle in die die Teiltabellen eingefuegt werden.
     * @param held
     *            dessen Werte in der Tabelle dargestellt werden.
     * @param zauber
     *            die in der Tabelle dargestellt werden.
     * @param zeile
     *            der Haupttabelle ab der die Teiltabelle eingefuegt wird. Die
     *            erste Zeile ist Zeile 0.
     * @param spalte
     *            der Haupttabelle in der die Teiltabellen eingefuegt wird. Die
     *            erste Spalte ist Spalte 0, die andere moegliche Spalte ist
     *            Spalte 6.
     * @return der Index des Zaubers, der nicht mehr in die Tabelle eingefuegt
     *         wurde.
     * @throws BadElementException
     */
    private void einfuegenZauber(
            final Table tabelle,
            final Held held,
            final ZauberTabelleDaten zauber,
            int zeile,
            int spalte ) throws BadElementException {

        while ( zauber.hatZaubergruppe() ) {

            final int freieZeilen = ZEILEN_PRO_SEITE - zeile;

            if ( 2 > freieZeilen ) {
                // Es sind nicht mehr genug freie Zeilen da um eine neue
                // Teiltabelle zu beginnen.
                if ( 0 == spalte ) {
                    // Die linke Haelfte der Haupttabelle ist gefuellt.
                    // Es kann jetzt die rechte Haelfte gefuellt werden.
                    spalte = 7;
                    zeile = 0;

                } else {

                    // Die ganze Haupttabelle ist voll.
                    // Es kann abgebrochen werden.
                    return;

                }
            }

            final Zaubergruppe zaubergruppe = zauber.getZaubergruppe();

            Cell zelle;

            zelle = erzeugeZelle(
                    zaubergruppe.getName(),
                    schrift12b,
                    Element.ALIGN_LEFT );

            zelle.setColspan( 4 );

            tabelle.addCell( zelle, zeile, spalte );

            zelle = erzeugeZelle(
                    ZFW,
                    schrift12b,
                    Element.ALIGN_CENTER );

            zelle.setColspan( 2 );

            tabelle.addCell( zelle );

            ++zeile;

            while ( zauber.hatNaechstenZauber() ) {

                final ZauberWert wert = zauber.getNaechstenZauber();

                final Zauber z = wert.getZauber();

                final StringBuilder text = new StringBuilder();

                text.append( wert.getName() );


                if ( held.hat( VorNachteil.ELFISCHE_WELTSICHT )
                  && held.hatHauszauber( z ) ) {

                    text.append( "** " );

                } else if ( held.hat( VorNachteil.ELFISCHE_WELTSICHT )
                         && held.hatLeitzauber( z ) ) {

                    text.append( "* " );

                } else {

                    text.append( " " );

                }

                text.append( wert.getProbe().toString() );

                zelle = erzeugeZelle(
                        text.toString(),
                        schrift10,
                        Element.ALIGN_LEFT );

                setRahmen( zelle, 0.1F, 0F, 0.1F, 0.1F );

                zelle.setColspan( 3 );

                tabelle.addCell( zelle, zeile, spalte );

                final int faktor = wert.getSteigerungsfaktor( held );

                zelle = erzeugeZelle( SKT.toString( faktor ),
                                      schrift10,
                                      Element.ALIGN_CENTER );

                setRahmen( zelle, 0.1F, 0.1F, 0.1F, 0F );

                tabelle.addCell( zelle );

                final String stufe = String.valueOf( wert.getStufe() );

                zelle = erzeugeZelle( stufe, schrift10, Element.ALIGN_RIGHT );

                tabelle.addCell( zelle );

                zelle = erzeugeZelle( " ", schrift10, Element.ALIGN_RIGHT );

                tabelle.addCell( zelle );

                ++zeile;

                if ( ZEILEN_PRO_SEITE == zeile ) {
                    // Eine Haelfte der Tabelle ist komplett gefuellt
                    // vielleicht auch beide.

                    if ( 0 == spalte ) {
                        // Die linke Haelfte der Haupttabelle ist gefuellt, jetzt
                        // wird die rechte Seite gefuellt.
                        spalte = 7;
                        zeile = 0;
                        break;

                    } else {

                        if ( ! zauber.hatNaechstenZauber() ) {
                            zauber.naechsteZaubergruppe();
                        }
                        // Die Tabelle ist komplett gefuellt.
                        // Es kann abgebrochen werden
                        return;
                    }
                }
            }

            if ( ! zauber.hatNaechstenZauber() ) {
                // Die innere Schleife hat alle Zauber der Gruppe
                // abgearbeitet.
                zauber.naechsteZaubergruppe();

                if ( 0 != zeile ) {
                    // Es ist nicht der Anfang der Haupttabelle.
                    // Es kommt eine Teiltabelle vorher.
                    // Zu dieser wird ein Leerzeile Platz gelassen.
                    ++zeile;
                }
            }
        }

    }

    /**
     * Befuellt zwei nebeneinander liegende Tabellenzellen mit Text.
     * Der Text der linken Spalte wird in fetter Schrift dargestellt, der Text
     * in der rechten Spalte in normaler Schrift.
     *
     * @param tabelle in welche die Spalten eingefuegt werden sollen.
     * @param text1 fuer die linke Spalte. Es kann auch ein 
     *          Lokalisierungsschluessel fuer einen Text sein. 
     * @param text2 fier die rechte Spalte. Es kann auch ein 
     *          Lokalisierungsschluessel fuer einen Text sein.
     * @param zeile in der der Text stehen soll.
     * @throws BadElementException
     */
    private void einfuegenZweiSpalten(
            final Table tabelle,
            String text1,
            String text2,
            final int zeile,
            final int spalte ) throws BadElementException {

        text1 = I18n.i18nVersuch( text1 );
        text2 = I18n.i18nVersuch( text2 );
        
        Cell cell = erzeugeZelle( text1, schrift12b, Element.ALIGN_LEFT );
        tabelle.addCell( cell, zeile, spalte );

        cell = erzeugeZelle( text2, schrift12, Element.ALIGN_LEFT );
        tabelle.addCell( cell, zeile, spalte+1 );
    }

    /**
     * @param held dessen Ausweichenwert in der Tabelle angezeigt werden
     *        sollen.
     * @return die Tabelle fuer den Ausweichenwert des Helden.
     * @throws BadElementException
     */
    private Table erzeugeAusweichenTabelle( final Held held )
    throws BadElementException {

        final Table tabelle = erzeugeTabelle( 10F, 2F, 5F, 2F, 43F, 2F, 10F );

        Cell zelle = erzeugeZelle(
                Basiswert.PAB.getKuerzel(),
                schrift12b,
                Element.ALIGN_LEFT );
        setRahmen( zelle, 0.1F, 0F, 0F, 1F );
        tabelle.addCell( zelle );

        zelle = new Cell();
        setRahmen( zelle, 0.1F, 0F, 0F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle( 
                BE, 
                schrift12b, 
                Element.ALIGN_CENTER );
        setRahmen( zelle, 0.1F, 0F, 0F, 0F );
        tabelle.addCell( zelle );

        zelle = new Cell();
        setRahmen( zelle, 0.1F, 0F, 0F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                SONDERFERTIGKEIT,
                schrift12b,
                Element.ALIGN_CENTER );
        setRahmen( zelle, 0.1F, 0F, 0F, 0F );
        tabelle.addCell( zelle );

        zelle = new Cell();
        setRahmen( zelle, 0.1F, 0F, 0F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                AUSWEICHEN,
                schrift12b,
                Element.ALIGN_CENTER );
        setRahmen( zelle, 0.1F, 0.1F, 0F, 0F );
        tabelle.addCell( zelle );

        // Zweite Zeile
        // ************
        zelle = erzeugeZelle(
                String.valueOf( held.getStufe( Basiswert.PAB ) ),
                schrift12,
                Element.ALIGN_CENTER );
        setRahmen( zelle, 0.0F, 0F, 0.1F, 0.1F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle( "-", schrift12b, Element.ALIGN_CENTER );
        setRahmen( zelle, 0F, 0F, 0.1F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                String.valueOf( held.getBehinderung() ),
                schrift12,
                Element.ALIGN_CENTER );
        setRahmen( zelle, 0F, 0F, 0.1F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle( "+", schrift12b, Element.ALIGN_CENTER );
        setRahmen( zelle, 0F, 0F, 0.1F, 0F );
        tabelle.addCell( zelle );

        final List< Sonderfertigkeit > ausweichen =
            new ArrayList< Sonderfertigkeit >();

        for ( final Sonderfertigkeit s : new Sonderfertigkeit[]{
                Sonderfertigkeit.AUSWEICHEN_I,
                Sonderfertigkeit.AUSWEICHEN_II,
                Sonderfertigkeit.AUSWEICHEN_III } ) {

            if ( held.hat( s ) ) {

                ausweichen.add( s );
            }
        }
        zelle = erzeugeZelle(
                Strings.erzeugeAufzaehlung( ausweichen ),
                schrift12,
                Element.ALIGN_LEFT );
        setRahmen( zelle, 0F, 0F, 0.1F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle( "=", schrift12b, Element.ALIGN_CENTER );
        setRahmen( zelle, 0F, 0F, 0.1F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                String.valueOf( held.getAusweichen() ),
                schrift12b,
                Element.ALIGN_CENTER );
        setRahmen( zelle, 0F, 0.1F, 0.1F, 0F );
        tabelle.addCell( zelle );

        return tabelle;
    }

    private Table erzeugeEigenschaftenBasiswerte( final Held held )
            throws DocumentException {

        final Table tabelle =
            erzeugeTabelle( 17F, 6F, 7F, 6F, 6F, 2F, 25F, 6F, 6F, 6F, 7F, 7F );

        // Linken Tabellenteil fuer die Eigenschaften erzeugen
        Cell zelle = erzeugeZelle( "", schrift6b, Element.ALIGN_LEFT );
        tabelle.addCell( zelle, 0, 0 );

        zelle = erzeugeZelle(
                EIGENSCHAFT_STARTWERT,
                schrift6b,
                Element.ALIGN_CENTER );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                EIGENSCHAFT_ZUKAUF,
                schrift6b,
                Element.ALIGN_CENTER );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                EIGENSCHAFT_AKTUELL,
                schrift6b,
                Element.ALIGN_CENTER );
        zelle.setColspan( 2 );
        tabelle.addCell( zelle );

        tabelle.addCell( erzeugeLeereZelleOhneRahmen() );

        final Eigenschaft[] eigenschaften = Eigenschaft.getEigenschaften();


        int z; // Die Zeile
        for ( z = 0; z < eigenschaften.length ; z++ ) {

            tabelle.addCell( erzeugeZelle(
                    eigenschaften[ z ].toString(),
                    schrift12,
                    Element.ALIGN_LEFT ), ( z + 1 ), 0 );

            final EigenschaftWert wert = held.getWert( eigenschaften[ z ] );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf( wert.getStartstufe() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf( wert.getZukauf() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf( wert.getStufe() ),
                    schrift12,
                    Element.ALIGN_RIGHT ) );

            tabelle.addCell( erzeugeZelle(
                    "",
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeLeereZelleOhneRahmen() );
        }

        tabelle.addCell( erzeugeZelle(
                GESCHWINDIGKEIT,
                schrift12,
                Element.ALIGN_LEFT ), ( z + 1 ), 0 );

        tabelle.addCell( erzeugeZelle(
                " ",
                schrift12,
                Element.ALIGN_RIGHT ) );

        tabelle.addCell( erzeugeZelle(
                " ",
                schrift12,
                Element.ALIGN_RIGHT ) );

        tabelle.addCell( erzeugeZelle(
                String.valueOf(  held.getGeschwindigkeit() ),
                schrift12,
                Element.ALIGN_RIGHT ) );

        tabelle.addCell( erzeugeZelle(
                " ",
                schrift12,
                Element.ALIGN_RIGHT ) );

        // ***********
        // BASISWERTE
        // **********
        tabelle.addCell( erzeugeZelle(
                "",
                schrift6b,
                Element.ALIGN_CENTER ), 0, 6 );

        tabelle.addCell( erzeugeZelle(
                BASISWERT_MOD,
                schrift6b,
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle(
                BASISWERT_BASISSTUFE,
                schrift6b,
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle(
                BASISWERT_AKTUELL,
                schrift6b,
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle(
                BASISWERT_ZUGEKAUFT,
                schrift6b,
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle(
                BASISWERT_MAX_ZUGEKAUFT,
                schrift6b,
                Element.ALIGN_CENTER ) );

        final Basiswert[] basiswerte = Basiswert.getBasiswerte();

        for ( z = 0 ; z < basiswerte.length ; ++z ) {

            final BasiswertWert wert = held.getWert( basiswerte[ z ] );

            if ( basiswerte[ z ].equals( Basiswert.AE )
              && ( ! held.istZauberer() ) ) {
                // Wenn der Held kein Zauberer ist, dann hat er auch keine
                // Astralenergie.
                tabelle.addCell( erzeugeZelle(
                        wert.getName(),
                        schrift12,
                        Element.ALIGN_LEFT ), z+1, 6 );

                // Der Rest der Zeile wird mit "-" gefuellt.
                zelle = erzeugeZelle(
                        "-",
                        schrift12,
                        Element.ALIGN_CENTER );

                tabelle.addCell( zelle );
                tabelle.addCell( zelle );
                tabelle.addCell( zelle );
                tabelle.addCell( zelle );
                tabelle.addCell( zelle );

            } else {
                // Es wird ganz normal die Zeile fuer den Basiswert ausgegeben.

                tabelle.addCell( erzeugeZelle(
                        wert.getName(),
                        schrift12,
                        Element.ALIGN_LEFT ), z+1, 6 );

                tabelle.addCell( erzeugeZelle(
                        String.valueOf( wert.getModifikation() ),
                        schrift12,
                        Element.ALIGN_CENTER ) );

                tabelle.addCell( erzeugeZelle(
                        String.valueOf( wert.getBasisstufe() ),
                        schrift12,
                        Element.ALIGN_CENTER ) );

                tabelle.addCell( erzeugeZelle(
                        String.valueOf( held.getStufe( basiswerte[ z ] ) ),
                        schrift12,
                        Element.ALIGN_CENTER ) );

                if ( z < 4 ) {
                    // Nur fuer LE, AU, AE und MR koennen Punkte zugekauft
                    // werden.

                    tabelle.addCell( erzeugeZelle(
                            String.valueOf(  wert.getZukauf() ),
                            schrift12,
                            Element.ALIGN_CENTER ) );

                    tabelle.addCell( erzeugeZelle(
                            String.valueOf( wert.getMaxZukauf() ),
                            schrift12,
                            Element.ALIGN_CENTER ) );
                }
            }
        }

        return tabelle;
    }

    /**
     * @param held dessen Waffen in die Tabelle eingefuegt werden sollen.
     * @return die Tabelle fuer die Fernkampfwaffen des Helden.
     * @throws BadElementException
     */
    private Table erzeugeFernkampfwaffentabelle( final Held held )
    throws BadElementException {

        final Table tabelle =
            erzeugeTabelle( 20F, 10F, 20F, 16F, 6F, 8F,
                            3F, 3F, 3F, 3F, 3F );
                            // 91
        tabelle.addCell( erzeugeZelle( 
                FERNKAMPFWAFFE, 
                schrift12b, 
                Element.ALIGN_LEFT ) );

        tabelle.addCell( erzeugeZelle( 
                TYP_EBE,
                schrift11b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                ENTFERNUNGEN,
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                TP_ENTFERNUNG,
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                FK,
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                TP,
                schrift12b, 
                Element.ALIGN_CENTER ) );

        final Cell zelle = erzeugeZelle( 
                GESCHOSSE, 
                schrift12b, 
                Element.ALIGN_CENTER );
        zelle.setColspan( 5 );
        tabelle.addCell( zelle );

        int i = 0;

        final List< FernkampfwaffeWert > waffen = held.getFernkampfwaffen();

        for ( final int n = waffen.size() ; i < n ; ++i ) {

            final FernkampfwaffeWert waffe = waffen.get( i );

            tabelle.addCell( erzeugeZelle( waffe.getName(),
                                           schrift12,
                                           Element.ALIGN_LEFT ) );

            tabelle.addCell( erzeugeZelle(
                    String.format( 
                            "%1$s%2$+d", 
                            I18n.i18n( BE ), 
                            -waffe.getTalent().getEBe() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    Strings.erzeugeAufzaehlung( waffe.getReichweiten() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    Strings.erzeugeAufzaehlung( waffe.getTpPlus() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf(  waffe.getFernkampfwert( held ) ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    waffe.getTp().toString(),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            for ( int j = 0 ; j < 5 ; ++j ) {

                tabelle.addCell( new Cell() );
            }
        }

        for ( ; i < Held.ANZAHL_FERNKAMPFWAFFEN ; ++i ) {
            for ( int j = 0 ; j < 11 ; ++j ) {
                tabelle.addCell( new Cell() );
            }
        }

        return tabelle;
    }

    /**
     * @param held dessen Initiativewert in der Tabelle angezeigt werden
     *        sollen.
     * @return die Tabelle fuer den Initiativewert des Helden.
     * @throws BadElementException
     */
    private Table erzeugeInitiativeTabelle( final Held held )
    throws BadElementException {

        final Table tabelle = erzeugeTabelle( 10F, 2F, 5F, 2F, 43F, 2F, 10F );

        Cell zelle = erzeugeZelle(
                Basiswert.INI.getKuerzel(),
                schrift12b,
                Element.ALIGN_LEFT );
        setRahmen( zelle, 0.1F, 0F, 0F, 1F );
        tabelle.addCell( zelle );

        zelle = new Cell();
        setRahmen( zelle, 0.1F, 0F, 0F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle( 
                BE,
                schrift12b, 
                Element.ALIGN_CENTER );
        setRahmen( zelle, 0.1F, 0F, 0F, 0F );
        tabelle.addCell( zelle );

        zelle = new Cell();
        setRahmen( zelle, 0.1F, 0F, 0F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                SONDERFERTIGKEIT,
                schrift12b,
                Element.ALIGN_CENTER );
        setRahmen( zelle, 0.1F, 0F, 0F, 0F );
        tabelle.addCell( zelle );

        zelle = new Cell();
        setRahmen( zelle, 0.1F, 0F, 0F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                INITIATIVE,
                schrift12b,
                Element.ALIGN_CENTER );
        setRahmen( zelle, 0.1F, 0.1F, 0F, 0F );
        tabelle.addCell( zelle );

        // Zweite Zeile
        // ************
        zelle = erzeugeZelle(
                String.valueOf( held.getStufe( Basiswert.INI ) ),
                schrift12,
                Element.ALIGN_CENTER );
        setRahmen( zelle, 0.0F, 0F, 0.1F, 0.1F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle( "-", schrift12b, Element.ALIGN_CENTER );
        setRahmen( zelle, 0F, 0F, 0.1F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                String.valueOf( held.getBehinderung() ),
                schrift12,
                Element.ALIGN_CENTER );
        setRahmen( zelle, 0F, 0F, 0.1F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle( "+", schrift12b, Element.ALIGN_CENTER );
        setRahmen( zelle, 0F, 0F, 0.1F, 0F );
        tabelle.addCell( zelle );

        String kampfreflexe = "";
        if ( held.hat( Sonderfertigkeit.KAMPFREFLEXE ) ) {

            kampfreflexe = Sonderfertigkeit.KAMPFREFLEXE.getName();
        }

        zelle = erzeugeZelle(
                kampfreflexe,
                schrift12,
                Element.ALIGN_LEFT );
        setRahmen( zelle, 0F, 0F, 0.1F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle( "=", schrift12b, Element.ALIGN_CENTER );
        setRahmen( zelle, 0F, 0F, 0.1F, 0F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                String.valueOf( held.getInitiative() ),
                schrift12b,
                Element.ALIGN_CENTER );
        setRahmen( zelle, 0F, 0.1F, 0.1F, 0F );
        tabelle.addCell( zelle );

        return tabelle;
    }

    /**
     * @return eine Einzeilige Tabelle mit den Kampfbasiswerten.
     *         (AT, PA, FK, INI)
     */
    private Table erzeugeKampfBasiswerteTabelle( final Held held )
    throws BadElementException {

        final Table tabelle =
            erzeugeTabelle( 21F, 4F, 21F, 4F, 21F, 4F, 21F, 4F );

        for ( final Basiswert basiswert : new Basiswert[]{ Basiswert.ATB,
                                                           Basiswert.PAB,
                                                           Basiswert.FKB,
                                                           Basiswert.INI } ) {

            final String name = basiswert.getName();

            Cell zelle = erzeugeZelle(
                    name + ":",
                    schrift10b,
                    Element.ALIGN_RIGHT );

            setRahmen( zelle, 0, 0, 0, 0 );

            tabelle.addCell( zelle );

            final int stufe = held.getStufe( basiswert );

            final String stufeString = String.valueOf( stufe );

            zelle = erzeugeZelle( stufeString, schrift10, Element.ALIGN_LEFT );

            setRahmen( zelle, 0, 0, 0, 0 );

            tabelle.addCell( zelle );
        }

        return tabelle;
    }

    /**
     * Eine einzeilige Tabelle mit 16 Spalten die fuer alle 8 Eigenschaften,
     * das Kuerzel der Eigenschaft und die Stufe des Helden in der Eigenschaft
     * anzeigt.
     *
     * @param held dessen Stufen in den Eigenschaften angezeigt werden sollen.
     * @return eine Tabelle mit den Eigenschafts-Stufen des Helden.
     * @throws BadElementException
     */
    private Table erzeugeKleineEigenschaftenTabelle( final Held held )
    throws BadElementException {

        final Table tabelle = erzeugeTabelle(
                1F, 1F, 1F, 1F, 1F, 1F, 1F, 1F,
                1F, 1F, 1F, 1F, 1F, 1F, 1F, 1F  );

        tabelle.setBorder(
                Table.LEFT | Table.RIGHT | Table.TOP | Table.BOTTOM );

        for ( final Eigenschaft eigenschaft : Eigenschaft.getEigenschaften()) {

            final String kuerzel = eigenschaft.getKuerzel();

            Cell zelle = erzeugeZelle(
                    kuerzel + ":",
                    schrift12b,
                    Element.ALIGN_RIGHT );

            setRahmen( zelle, 0, 0, 0, 0 );

            tabelle.addCell( zelle );

            final int stufe = held.getStufe( eigenschaft );

            final String stufeString = String.valueOf( stufe );

            zelle = erzeugeZelle( stufeString, schrift12, Element.ALIGN_LEFT );

            setRahmen( zelle, 0, 0, 0, 0 );

            tabelle.addCell( zelle );
        }

        return tabelle;
    }

    /**
     * @return eine leere Tabellenzelle ohne Rahmen.
     * @throws BadElementException
     */
    private Cell erzeugeLeereZelleOhneRahmen() throws BadElementException {

        final Cell cell = new Cell();

        cell.setBorderWidth( 0F );

        return cell;
    }

    /**
     * @param held dessen Werte in Raufen und Rignen in die Tabelle eingefuegt
     *        werden sollen.
     * @return die Tabelle fuer den waffenlosen Nahkampf.
     * @throws BadElementException
     */
    private Table erzeugeNahkampftabelle( final Held held )
    throws BadElementException {

        final Table tabelle =
            erzeugeTabelle( 20F, 10F, 10F, 10F, 10F, 10F, 10F );

        tabelle.addCell( erzeugeZelle( 
                WAFFENLOSER_KAMPF,
                schrift12b, 
                Element.ALIGN_LEFT ) );

        tabelle.addCell( erzeugeZelle( 
                TYP_EBE,
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                TP_KK,
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                INI,
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                AT,
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                PA,
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                TP_A,
                schrift12b, 
                Element.ALIGN_CENTER ) );

        // RAUFEN
        // ******
        Kampftechnik talent = Talent.RAUFEN;

        tabelle.addCell( erzeugeZelle( talent.getName(),
                                       schrift12,
                                       Element.ALIGN_LEFT ) );

        tabelle.addCell( erzeugeZelle(
                String.format( "%1$s%2$+d", I18n.i18n( BE ), talent.getEBe() ),
                schrift12,
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( "10/3",
                                       schrift12,
                                       Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( "+0",
                                       schrift12,
                                       Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle(
                String.valueOf( held.getAttackeRaufen() ),
                schrift12,
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle(
                String.valueOf( held.getParadeRaufen() ),
                schrift12,
                Element.ALIGN_CENTER ) );

        final Tp tp = new Tp( "W" );

        final int kk = held.getStufe( Eigenschaft.KK );

        final int tpPlus = ( kk - 10 ) / 3;

        tabelle.addCell( erzeugeZelle( tp.getTpFormel( tpPlus ),
                                       schrift12,
                                       Element.ALIGN_CENTER ) );

        // RINGEN
        // ******
        talent = Talent.RINGEN;

        tabelle.addCell( erzeugeZelle( talent.getName(),
                                       schrift12,
                                       Element.ALIGN_LEFT ) );

        tabelle.addCell( erzeugeZelle(
                String.format( "%1$s%2$+d", I18n.i18n( BE ), talent.getEBe() ),
                schrift12,
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( "10/3",
                                       schrift12,
                                       Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( "+0",
                                       schrift12,
                                       Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle(
                String.valueOf( held.getAttackeRingen() ),
                schrift12,
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle(
                String.valueOf( held.getParadeRingen() ),
                schrift12,
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( tp.getTpFormel( tpPlus ),
                                       schrift12,
                                       Element.ALIGN_CENTER ) );

        return tabelle;
    }

    /**
     * @param held dessen Waffen in die Tabelle eingefuegt werden sollen.
     * @return die Tabelle fuer die Nahkampfwaffen des Helden.
     * @throws BadElementException
     */
    private Table erzeugeNahkampfwaffentabelle( final Held held )
    throws BadElementException {

        final Table tabelle =
            erzeugeTabelle( 20F, 10F, 6F, 8F, 8F, 5F, 6F, 5F, 5F, 8F,
                            3F, 3F, 3F, 3F, 3F );
                            // 91
        tabelle.addCell( erzeugeZelle( 
                NAHKAMPFWAFFE, 
                schrift12b, 
                Element.ALIGN_LEFT ) );

        tabelle.addCell( erzeugeZelle( 
                TYP_EBE, 
                schrift11b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                DK, 
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                TP, 
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                TP_KK, 
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                INI, 
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                WM, 
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                AT, 
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                PA, 
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                TP, 
                schrift12b, 
                Element.ALIGN_CENTER ) );

        final Cell zelle = erzeugeZelle( 
                BRUCHFAKTOR, 
                schrift12b, 
                Element.ALIGN_CENTER );
        zelle.setColspan( 5 );
        tabelle.addCell( zelle );

        int i = 0;

        final List< NahkampfwaffeWert > waffen = held.getNahkampfwaffen();

        for ( final int n = waffen.size() ; i < n ; ++i ) {

            final NahkampfwaffeWert waffe = waffen.get( i );

            tabelle.addCell( erzeugeZelle( waffe.getName(),
                                           schrift12,
                                           Element.ALIGN_LEFT ) );

            tabelle.addCell( erzeugeZelle(
                    String.format( 
                            "%1$s%2$+d", 
                            I18n.i18n( BE ),
                            -waffe.getTalent().getEBe() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle( waffe.getDkString(),
                                           schrift12,
                                           Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle( waffe.getTp().toString(),
                                           schrift12,
                                           Element.ALIGN_CENTER ) );

            final String tpKk = String.format( "%1$s/%2$s",
                                               waffe.getSchwellenwert(),
                                               waffe.getSchadensschritt() );

            tabelle.addCell( erzeugeZelle( tpKk,
                                           schrift12,
                                           Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf( waffe.getIni() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            final String wm = String.format( "%1$s/%2$s",
                                             waffe.getWmAt(),
                                             waffe.getWmPa() );

            tabelle.addCell( erzeugeZelle( wm,
                                           schrift12,
                                           Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf( waffe.getAttacke( held ) ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf( waffe.getParade( held ) ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf( waffe.getTp( held ) ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf( waffe.getBf() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            for ( int j = 0 ; j < 4 ; ++j ) {

                tabelle.addCell( new Cell() );
            }
        }

        for ( ; i < Held.ANZAHL_NAHKAMPFWAFFEN ; ++i ) {
            for ( int j = 0 ; j < 15 ; ++j ) {
                tabelle.addCell( new Cell() );
            }
        }

        return tabelle;
    }

    /**
     * Erzeugt einen Paragraph mit einer Uebeschrift.
     *
     * @param titel des Paragraphs. Es kann auch ein Lokalisierungsschluessel
     *          fuer den Titel uebergeben werden.
     * @return der erzeugte Paragraph
     */
    private final Paragraph erzeugeParagraph ( String titel ) {

        titel = I18n.i18nVersuch( titel );
        
        final Paragraph paragraph = new Paragraph( titel, schrift16b );

        paragraph.setAlignment( Paragraph.ALIGN_CENTER );

        return paragraph;
    }

    /**
     * Erzeugt die Tabelle mit den persoenlichen Angaben des Helden wie
     * Aussehen, Groesse usw. und den Informationen ueber das Umfeld des
     * Helde, wie Herkunft und Stand.
     *
     * @param held dessen Daten dargestellt werden sollen.
     * @return die fertige Tabelle
     * @throws DocumentException
     */
    private Table erzeugePersoenlichesUmfeld(
            final Held held )
            throws DocumentException {

        final int apGesamt = held.getApGesamt();

        final int apGuthaben = held.getApGuthaben();

        final int apEingesetzt = held.getApEingesetzt();

        final Table tabelle = erzeugeTabelle( 19F, 30F, 2F, 19F, 30F );

        int zeile = 0;
        /* *****************
            linke Spalte
        ****************** */
        einfuegenZweiSpalten(
                tabelle,
                GESCHLECHT,
                held.getGeschlecht().toString(),
                zeile, 0 );
        tabelle.addCell( erzeugeLeereZelleOhneRahmen(), zeile, 2 );

        ++zeile;
        einfuegenZweiSpalten(
                tabelle,
                GEBURTSTAG,
                held.getGeburtstag() + " / " + held.getAlter(),
                zeile, 0 );
        tabelle.addCell( erzeugeLeereZelleOhneRahmen(), zeile, 2 );

        ++zeile;
        einfuegenZweiSpalten(
                tabelle,
                GROESSE,
                held.getGroesse() + " " + I18n.i18n( HALBFINGER ),
                zeile, 0 );
        tabelle.addCell( erzeugeLeereZelleOhneRahmen(), zeile, 2 );

        ++zeile;
        einfuegenZweiSpalten(
                tabelle,
                GEWICHT,
                held.getGewicht() + " " + I18n.i18n( STEIN ),
                zeile, 0 );
        tabelle.addCell( erzeugeLeereZelleOhneRahmen(), zeile, 2 );

        ++zeile;
        einfuegenZweiSpalten(
                tabelle,
                HAARFARBE,
                held.getHaarfarbe().getName(),
                zeile, 0 );
        tabelle.addCell( erzeugeLeereZelleOhneRahmen(), zeile, 2 );

        ++zeile;
        einfuegenZweiSpalten(
                tabelle,
                AUGENFARBE,
                held.getAugenfarbe().getName(),
                zeile, 0 );
        tabelle.addCell( erzeugeLeereZelleOhneRahmen(), zeile, 2 );

        ++zeile;
        einfuegenZweiSpalten(
                tabelle,
                AUSSEHEN,
                held.getAussehen(),
                zeile, 0 );
        tabelle.addCell( erzeugeLeereZelleOhneRahmen(), zeile, 2 );

        /* *****************
            rechte Spalte
         ****************** */
        zeile = 0;

        einfuegenZweiSpalten(
                tabelle,
                STAND,
                held.getStand(),
                zeile, 3 );

        ++zeile;
        einfuegenZweiSpalten(
                tabelle,
                TITEL,
                held.getTitel(),
                zeile, 3 );

        ++zeile;
        einfuegenZweiSpalten(
                tabelle,
                Eigenschaft.SO.getName(),
                String.valueOf( held.getStufe( Eigenschaft.SO ) ),
                zeile, 3 );

        ++zeile;
        einfuegenZweiSpalten(
                tabelle,
                GESAMT_AP,
                String.valueOf( apGesamt ),
                zeile, 3 );

        ++zeile;
        einfuegenZweiSpalten(
                tabelle,
                EINGESETZTE_AP,
                String.valueOf( apEingesetzt ),
                zeile, 3 );

        ++zeile;
        einfuegenZweiSpalten(
                tabelle,
                AP_GUTHABEN,
                String.valueOf( apGuthaben ),
                zeile, 3 );

        ++zeile;
        einfuegenZweiSpalten(
                tabelle,
                HINTERGRUND,
                held.getHintergrund(),
                zeile, 3 );

        return tabelle;
    }

    /**
     * Erzeugt die Tabelle mit den Informationen zu Rasse, Kultur und
     * Profession.
     *
     * @param held dessen Daten dargestellt werden sollen.
     * @return eine Tabelle mit Rasse, Kultur und Profession.
     * @throws DocumentException
     */
    private final Table erzeugePersonendaten ( final Held held )
            throws DocumentException {

        final Table tabelle = erzeugeTabelle( 19F, 81F );

        einfuegenZweiSpalten(
                tabelle,
                NAME,
                held.getName(),
                0, 0 );

        einfuegenZweiSpalten(
                tabelle,
                RASSE,
                held.getRasse().getName(),
                1, 0 );

        einfuegenZweiSpalten(
                tabelle,
                KULTUR,
                held.getKultur().getName(),
                2, 0 );

        einfuegenZweiSpalten(
                tabelle,
                PROFESSION,
                held.getProfession().getName( held ),
                3, 0 );

        return tabelle;
    }

    /**
     * @param held dessen Ruestungsstuecke in die Tabelle eingefuegt werden
     *        sollen.
     * @return die Tabelle fuer die Ruestungen des Helden.
     * @throws BadElementException
     */
    private Table erzeugeRuestungstabelle( final Held held )
    throws BadElementException {

        final Table tabelle = erzeugeTabelle( 20F, 5F, 5F, 45F );

        tabelle.addCell( erzeugeZelle( 
                RUESTUNGSSTUECK, 
                schrift12b, 
                Element.ALIGN_LEFT ) );

        tabelle.addCell( erzeugeZelle( 
                RS, 
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeZelle( 
                BE, 
                schrift12b, 
                Element.ALIGN_CENTER ) );

        tabelle.addCell( erzeugeLeereZelleOhneRahmen() );

        int i = 0;

        final List< RuestungWert > ruestungen = held.getRuestungen();

        for ( final int n = ruestungen.size() ; i < n ; ++i ) {

            final RuestungWert ruestung = ruestungen.get( i );

            tabelle.addCell( erzeugeZelle( ruestung.getName(),
                                           schrift12,
                                           Element.ALIGN_LEFT ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf( ruestung.getRs() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf( ruestung.getBe() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeLeereZelleOhneRahmen() );
        }

        for ( ; i < Held.ANZAHL_RUESTUNGSTEILE; ++i ) {
            for ( int j = 0 ; j < 3 ; ++j ) {
                tabelle.addCell( new Cell() );
            }
            tabelle.addCell( erzeugeLeereZelleOhneRahmen() );
        }

        Cell zelle = erzeugeZelle ( SUMME, schrift12b, Element.ALIGN_LEFT);
        zelle.setBorderWidthTop( 0.2F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle (
                String.valueOf( held.getRuestungsschutz() ),
                schrift12b,
                Element.ALIGN_CENTER );
        zelle.setBorderWidthTop( 1F );
        tabelle.addCell( zelle );

        zelle = erzeugeZelle(
                String.valueOf( held.getBehinderung() ),
                schrift12b,
                Element.ALIGN_CENTER );
        zelle.setBorderWidthTop( 1F );
        tabelle.addCell( zelle );

        tabelle.addCell( erzeugeLeereZelleOhneRahmen() );

        return tabelle;
    }

    /**
     * @param held dessen Schilde in die Tabelle eingefuegt werden sollen.
     * @return die Tabelle fuer die Schilde und Parierwaffen des Helden.
     * @throws BadElementException
     */
    private Table erzeugeSchildetabelle( final Held held )
    throws BadElementException {

        final Table tabelle =
            erzeugeTabelle( 20F, 10F, 10F, 10F, 10F,
                            3F, 3F, 3F, 3F, 3F );
                            // 5
        tabelle.addCell(
                erzeugeZelle( NAME, schrift12b, Element.ALIGN_LEFT ) );

        tabelle.addCell(
                erzeugeZelle( TYP, schrift12b, Element.ALIGN_CENTER ) );

        tabelle.addCell(
                erzeugeZelle( INI, schrift12b, Element.ALIGN_CENTER ) );

        tabelle.addCell(
                erzeugeZelle( WM, schrift12b, Element.ALIGN_CENTER ) );

        tabelle.addCell(
                erzeugeZelle( PA, schrift12b, Element.ALIGN_CENTER ) );

        final Cell zelle =
            erzeugeZelle( BRUCHFAKTOR, schrift12b, Element.ALIGN_CENTER );
        zelle.setColspan( 5 );
        tabelle.addCell( zelle );

        int i = 0;

        final List< SchildWert > schilde = held.getSchilde();

        for ( final int n = schilde.size() ; i < n ; ++i ) {

            final SchildWert schild = schilde.get( i );

            tabelle.addCell( erzeugeZelle( schild.getName(),
                                           schrift12,
                                           Element.ALIGN_LEFT ) );

            tabelle.addCell( erzeugeZelle(
                    schild.getTyp().toString(),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf( schild.getIni() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.format( "%1$s/%2$s",
                                   schild.getWmAt(),
                                   schild.getWmPa() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf(  schild.getParade( held ) ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            tabelle.addCell( erzeugeZelle(
                    String.valueOf(  schild.getBf() ),
                    schrift12,
                    Element.ALIGN_CENTER ) );

            for ( int j = 0 ; j < 4 ; ++j ) {

                tabelle.addCell( new Cell() );
            }
        }

        for ( ; i < Held.ANZAHL_SCHILDE; ++i ) {
            for ( int j = 0 ; j < 10 ; ++j ) {
                tabelle.addCell( new Cell() );
            }
        }

        return tabelle;
    }

    /**
     * Erzeugt eine Tabelle mit einer vorgegebenen Spaltenzahl,
     * einem Padding von 5 und einer Breite von 100%.
     *
     * @param spaltenbreiten fuer die Spalten der Tabelle. Die Werte werden
     *      relativ zueinander betrachtet. Eine Angabe von {@code 2F, 1F, 1F }
     *      bedeutet dass die erste Spalte doppelt so breit ist wie die beiden
     *      andern.
     * @return die fertige Tabelle.
     * @throws BadElementException
     */
    private Table erzeugeTabelle( final float ... spaltenbreiten )
        throws BadElementException {

        final Table tabelle = new Table( spaltenbreiten.length );

        tabelle.setPadding( 5 );
        tabelle.setWidth( 100F );
        tabelle.setWidths( spaltenbreiten );
        tabelle.setBorderWidth( 0F );

        return tabelle;
    }


    /**
     * Erzeugt eine Tabellenzelle mit einem bestimmten Text, Schriftart und
     * Textausrichtung.
     *
     * @param text der in der Zelle stehen soll. Es kann auch ein Schluessel
     *          fuer einen lokalisierten Text uebergeben werden.
     * @param schriftart die fuer den Text verwendet werden soll.
     * @param ausrichtung des Textes in der Zelle.
     * @return die fertige Tabellenzelle.
     * @throws BadElementException
     */
    private Cell erzeugeZelle(
            String text,
            final Font schriftart,
            final int ausrichtung ) throws BadElementException {

        text = I18n.i18nVersuch( text );
        
        final Paragraph paragraph = new Paragraph( text, schriftart );

        paragraph.setAlignment( ausrichtung );

        final Cell cell = new Cell( paragraph );

        return cell;
    }

    /**
     * Bestimmt den Rahmen einer Tabellenzelle.
     *
     * @param zelle deren Rahmen gesetzt werden soll
     * @param oben Die dicke des Rahmens am oberen Rand der Zelle.
     * @param rechts Die dicke des Rahmens am rechten Rand der Zelle.
     * @param unten Die dicke des Rahmens am unteren Rand der Zelle.
     * @param links Die dicke des Rahmens am linken Rand der Zelle.
     */
    private void setRahmen(
            final Cell zelle,
            final float oben,
            final float rechts ,
            final float unten,
            final float links ) {

        zelle.setBorderWidthTop   ( oben   );
        zelle.setBorderWidthRight ( rechts );
        zelle.setBorderWidthBottom( unten  );
        zelle.setBorderWidthLeft  ( links  );
    }

    /**
     * Speichert ein Heldendokument eines Helden in eine bestimmte Datei.
     *
     * @param held dessen Dokument erzeugt werden soll.
     * @param datei in die das Dokument gespeichert werden soll.
     * @return {@code true}, wenn die Datei erfolgreich gespeichert wurde.
     */
    public final boolean speichern(
            final Held held,
            final File datei ) {

        final Document document = new Document();

        try {
            RtfWriter2.getInstance( document, new FileOutputStream( datei ) );
            // PdfWriter.getInstance( document, new FileOutputStream( datei ) );
            // HtmlWriter.getInstance( document, new FileOutputStream( datei ) );

            document.open();

            Paragraph paragraph;

            paragraph = new Paragraph( I18n.i18n( HELDENDOKUMENT ), schrift22b);
            paragraph.setAlignment( Paragraph.ALIGN_CENTER );
            document.add( paragraph );

            // Personendaten (Name, Rasse, Kultur, Profession)
            // ***********************************************
            paragraph = new Paragraph();
            paragraph.add( erzeugePersonendaten( held ) );
            document.add( paragraph );

            // Persoenliches und Umfeld (Aussehn, Titel, Herkunft ...)
            // *******************************************************
            paragraph = new Paragraph();
            paragraph.add( erzeugePersoenlichesUmfeld( held ) );
            document.add( paragraph );

            // Eigenschaften & Basiswerte
            // **************************
            paragraph = erzeugeParagraph( EIGENSCHAFTEN_BASISWERTE );
            paragraph.add( erzeugeEigenschaftenBasiswerte( held ) );
            document.add( paragraph );

            // Sonderfertigkeiten
            // ******************
            paragraph = erzeugeParagraph( SONDERFERTIGKEITEN );
            Table tabelle = erzeugeTabelle( 100F );
            List< SonderfertigkeitWert > sfen = held.getSonderfertigkeiten();
            sfen.sort( SonderfertigkeitWert.NAME_COMPARATOR );
            final String sonderfertigkeiten = Strings.erzeugeAufzaehlung( sfen );
            Cell zelle = erzeugeZelle(
                    sonderfertigkeiten,
                    schrift12,
                    Element.ALIGN_LEFT );
            tabelle.addCell( zelle );
            paragraph.add( tabelle );
            document.add( paragraph );

            // Vor-/Nachteile
            // **************
            paragraph = erzeugeParagraph( VOR_NACHTEILE );
            tabelle = erzeugeTabelle( 100F );
            List< VorNachteilWert > vne = held.getVorNachteile();
            vne.sort( VorNachteilWert.NAME_COMPARATOR );
            final String vorNachteile = Strings.erzeugeAufzaehlung( vne );
            zelle = erzeugeZelle(
                    vorNachteile,
                    schrift12,
                    Element.ALIGN_LEFT );
            tabelle.addCell( zelle );
            paragraph.add( tabelle );
            document.add( paragraph );

            // Talente
            // *******
            document.newPage();

            document.add( erzeugeKleineEigenschaftenTabelle( held ) );

            paragraph = erzeugeParagraph( TALENTE );

            tabelle = erzeugeTalenteTabelle();

            int zeile = einfuegenKampftechniken( tabelle, held );
            // eine Leerzeile lassen.
            ++zeile;

            final TalenteTabelleDaten talente;
            talente = new TalenteTabelleDaten( held );

            Tabellenende ende = new Tabellenende( 0, 0 );

            while ( talente.hatTalentgruppe() ) {

                ende = einfuegenTalente( tabelle, held, talente, zeile );

                if ( talente.hatTalentgruppe() ) {
                    // Eine Tabelle ist voll geschrieben worden, es sind aber
                    // noch Talente ueber.
                    paragraph.add( tabelle );

                    document.add( paragraph );

                    // Es haben nicht alle Talente in letzte Tabelle gepasst.
                    // Es muss noch eine weitere Seite mit einer weiteren
                    // Tabelle angelegt werden.
                    document.newPage();

                    document.add( erzeugeKleineEigenschaftenTabelle( held ) );

                    paragraph = erzeugeParagraph( TALENTE );

                    tabelle = erzeugeTalenteTabelle();

                    zeile = 0;
                }
            }

            // Zauber
            // *******
            final ZauberTabelleDaten zauber;
            zauber = new ZauberTabelleDaten( held );

            while ( zauber.hatZaubergruppe() ) {

                einfuegenZauber( tabelle,
                                 held,
                                 zauber,
                                 ende.zeile,
                                 ende.spalte );

                if ( zauber.hatZaubergruppe() ) {

                    paragraph.add( tabelle );

                    document.add( paragraph );

                    // Es haben nicht alle Zauber in letzte Tabelle gepasst.
                    // Es muss noch eine weitere Seite mit einer weiteren
                    // Tabelle angelegt werden.
                    document.newPage();

                    document.add( erzeugeKleineEigenschaftenTabelle( held ) );

                    paragraph = erzeugeParagraph( TALENTE );

                    tabelle = erzeugeTalenteTabelle();

                    ende = new Tabellenende( 0, 0 );
                }
            }

            paragraph.add( tabelle );

            document.add( paragraph );

            // WAFFEN
            // ******
            document.newPage();

            paragraph = new Paragraph();
            paragraph.add( erzeugeKampfBasiswerteTabelle( held ) );
            document.add( paragraph );

            paragraph = erzeugeParagraph( WAFFEN_KAMPFWERTE );
            paragraph.add( erzeugeNahkampfwaffentabelle( held ) );
            paragraph.add( erzeugeFernkampfwaffentabelle( held ) );
            paragraph.add( erzeugeNahkampftabelle( held ) );
            document.add( paragraph );

            paragraph = erzeugeParagraph( SCHILD_PARIERWAFFE );
            paragraph.add( erzeugeSchildetabelle( held ) );
            document.add( paragraph );

            paragraph = erzeugeParagraph( RUESTUNG );
            paragraph.add( erzeugeRuestungstabelle( held ) );
            paragraph.add( erzeugeAusweichenTabelle( held ) );
            paragraph.add( erzeugeInitiativeTabelle( held ) );
            document.add( paragraph );

        } catch (final Exception e) {
            
            DialogFehlermeldung.zeigeDialog( e );
            
            return false;
            
        } finally {
            
            document.close();
        }
        
        return true;
    }

    /**
     * @return eine Tabelle in die die Talente eingefuegt werden koennen.
     *
     * @throws BadElementException
     */
    private Table erzeugeTalenteTabelle() throws BadElementException {
        Table tabelle;
        tabelle = erzeugeTabelle(
                25F, 4F, 5F, 5F, 5F, 5F, 2F, 25F, 4F, 5F, 5F, 5F, 5F );
        return tabelle;
    }

    /**
     * Gibt an an welcher Stelle in der Tabelle der naechste Eintrag gemacht
     * werden kann.
     */
    private static class Tabellenende {

        public final int spalte;

        public final int zeile;

        private Tabellenende( final int zeile, final int spalte ) {

            this.spalte = spalte;

            this.zeile = zeile;
        }
    }
}
