/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.bausteine;

import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_ID;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_LEITTALENTE;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_MODIFIKATIONEN;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_NACHTEIL_BONI;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_SONDERFERTIGKEIT_BONI;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_TALENT;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_TALENT_BONI;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_UNGEEIGNETE_VOR_NACH;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_VERBILLIGTE_VARIANTEN;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_VERBINDUNGEN;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_VERB_SONDERFERTIGKEITEN;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_VORAUSSETZUNGEN;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_VORTEIL_BONI;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_ZAUBER_BONI;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;
import de.bernhardjung.dsaprogramm.basis.ElementMitNameImpl;
import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.basis.KonfliktBeschreibung;
import de.bernhardjung.dsaprogramm.basiswerte.Basiswert;
import de.bernhardjung.dsaprogramm.bausteine.bonus.SonderfertigkeitBonus;
import de.bernhardjung.dsaprogramm.bausteine.bonus.Talentbonus;
import de.bernhardjung.dsaprogramm.bausteine.bonus.TalentbonusFest;
import de.bernhardjung.dsaprogramm.bausteine.bonus.TalentbonusFrei;
import de.bernhardjung.dsaprogramm.bausteine.bonus.VerbilligteSonderfertigkeit;
import de.bernhardjung.dsaprogramm.bausteine.bonus.VerbilligteVariante;
import de.bernhardjung.dsaprogramm.bausteine.bonus.VorNachteilBonus;
import de.bernhardjung.dsaprogramm.bausteine.bonus.Zauberbonus;
import de.bernhardjung.dsaprogramm.bausteine.kosten.Kosten;
import de.bernhardjung.dsaprogramm.bausteine.kosten.KostenFest;
import de.bernhardjung.dsaprogramm.bausteine.ungeeignete.UngeeigneterVorNachteil;
import de.bernhardjung.dsaprogramm.bausteine.util.BasiswertModifikation;
import de.bernhardjung.dsaprogramm.bausteine.util.EigenschaftsModifikation;
import de.bernhardjung.dsaprogramm.bausteine.voraussetzungen.Voraussetzung;
import de.bernhardjung.dsaprogramm.bausteine.voraussetzungen.VoraussetzungEigenschaft;
import de.bernhardjung.dsaprogramm.daten.factorys.bausteine.FactoryEmpfohleneVorNachteileOut;
import de.bernhardjung.dsaprogramm.eigenschaften.Eigenschaft;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.Sonderfertigkeit;
import de.bernhardjung.dsaprogramm.talente.Talent;
import de.bernhardjung.dsaprogramm.talente.Talentgruppe;
import de.bernhardjung.dsaprogramm.vornachteile.VorNachteil;
import de.bernhardjung.dsaprogramm.vornachteile.VorNachteilWert;

public abstract class Heldenbaustein
extends ElementMitNameImpl {

    /**
     * Die Boni, die der Baustein auf die Basiswerte vergibt.
     */
    private final
    Map< Basiswert, BasiswertModifikation > basiswertModifikationen;

    /**
     * Die Boni, die der Baustein auf die Eigenschaften gibt.
     */
    private final
    Map< Eigenschaft, EigenschaftsModifikation > eigenschaftsModifikationen;

    /**
     * Die Voraussetzungen fuer die Eigenschaften des Helden der diesen
     * Baustein waehlen will.
     */
    protected final
    Map< Eigenschaft, VoraussetzungEigenschaft > eigenschaftsVoraussetzungen;

    /**
     * Eine sortierte Liste der empfohlenen Nachteile.
     */
    private final List< VorNachteil > empfohleneNachteile;

    /**
     * Die empfohlenen Vor-/Nachteile. Die Speicherung als (Hash)Set erlaubt es
     * schnell zu ueberpruefen ob ein Vor-/Nachteil empfohlen ist.
     */
    private final Set< VorNachteil > empfohleneVorNachteile;

    /**
     * Eine sortierte Liste der empfohlenen Vorteile.
     */
    private final List< VorNachteil > empfohleneVorteile;

    /**
     * Die GP, die bei der Berechnung der Kosten des Heldenbausteins, am Ende
     * von der Summe abgezogen werden.
     */
    private final double gpKostenAbzug;

    /**
     * Die GP-Kosten die bei der Erschaffung fuer den Baustein zu zahlen sind.
     */
    private Kosten kosten;

    /**
     * Die Leittalente die der Baustein besitzt.
     * Die Collection enthaelt nicht alle Talente die durch den Baustein zu
     * Leittalenten ernannt werden. Die Talente die durch einen
     * {@link TalentbonusFest} zum Leittalent werden muessen hier nicht
     * enthalten sein. Es geht mehr um die Talente die Leittalente sind, die
     * aber nicht durch den Baustein aktiviert werden.
     * Zu Leittalente siehe Regelwerk S 94, Elfische Weltsicht und Kasten
     * Seite 82.
     */
    private final Collection< Talent > leittalente;

    /** Die Nachteile die der Held automatisch durch den Baustein erhaelt. */
    private final List< VorNachteilBonus > nachteile;

    /**
     * Der Name des Bausteins z.B. "Tulamiden" oder "Bornland".
     */
  //  private String name;

    /** Die Sonderfertigkeiten die ein Held durch den Baustein erhaelt. */
    private final List< SonderfertigkeitBonus > sonderfertigkeiten;

    /** Die Talentboni welche der Baustein vergibt. */
    private final
    Map< Talentgruppe, List< Talentbonus > > talentboni;

    /** Eine sortierte Liste der ungeeigneten Nachteile. */
    private final List< UngeeigneterVorNachteil > ungeeigneteNachteile;

    /**
     * Die ungeeigneten Vor-/Nachteile. Die Speicherung als (Hash)Set erlaubt es
     * schnell zu ueberpruefen ob ein Vor-/Nachteil als ungeeignet gilt.
     */
    private final
    Map< VorNachteil, UngeeigneterVorNachteil > ungeeigneteVorNachteile;

    /** Eine sortierte Liste der ungeeigneten Vorteile. */
    private final List< UngeeigneterVorNachteil > ungeeigneteVorteile;

    /**
     * Die Sonderfertigkeiten die ein Held durch den Baustein verbilligt
     * erhaelt.
     */
    private final
    List< VerbilligteSonderfertigkeit > verbilligteSonderfertigkeiten;

    /**
     * Die Sonderfertigkeiten die ein Held durch den Baustein verbilligt
     * erhaelt.
     */
    private final List< VerbilligteVariante > verbilligteVariante;

    /**
     * Die Punkte die auf den Vorteil Verbindungen verteilt werden duerfen.
     * Siehe Profession Streuner.
     */
    private int verbindungen;

    /**
     * Die Voraussetzungen die ein Held erfuellen muss wenn er diesen Baustein
     * waehlen will. Das beinhaltet auch die Voraussetzungen fuer die
     * Eigenschaften ({@link #eigenschaftsVoraussetzungen}).
     */
    private final List< Voraussetzung > voraussetzungen;

    /** Die Vorteile die der Held durch den Baustein erhaelt. */
    private final List< VorNachteilBonus > vorteile;

    /** Die Zauber, die man durch den Baustein erhaelt. */
    private final List< Zauberbonus > zauberboni;

    /**
     * Erzeugt einen neuen Heldenbaustein.
     *
     * @see Rasse
     * @see Kultur
     * @see Profession
     *
     * @param dieId des Bausteins.
     * @param derGpKostenAbzug
     *            die GP, die bei der Berechnung der Kosten des Bausteins, am
     *            Ende von der Summe abgezogen werden.
     */
    Heldenbaustein( 
            final String dieId,
            final double derGpKostenAbzug ) {
        
        super( dieId );
        
        basiswertModifikationen =
            new HashMap< Basiswert, BasiswertModifikation >();

        eigenschaftsModifikationen =
            new HashMap< Eigenschaft, EigenschaftsModifikation >();

        eigenschaftsVoraussetzungen =
            new HashMap< Eigenschaft, VoraussetzungEigenschaft >();

        empfohleneNachteile = new ArrayList< VorNachteil >();

        empfohleneVorNachteile = new HashSet< VorNachteil >();

        empfohleneVorteile = new ArrayList< VorNachteil >();

        gpKostenAbzug = derGpKostenAbzug;

        kosten = new KostenFest( 0 );

        leittalente = new ArrayList< Talent >();

        nachteile = new ArrayList< VorNachteilBonus >();

        sonderfertigkeiten = new ArrayList< SonderfertigkeitBonus >();

        talentboni=  new HashMap< Talentgruppe, List< Talentbonus > >();
        initTalentBoni();

        ungeeigneteNachteile = new ArrayList< UngeeigneterVorNachteil >();

        ungeeigneteVorNachteile =
            new HashMap< VorNachteil, UngeeigneterVorNachteil >();

        ungeeigneteVorteile = new ArrayList< UngeeigneterVorNachteil >();

        verbilligteVariante =
            new ArrayList< VerbilligteVariante >();

        verbilligteSonderfertigkeiten =
            new ArrayList< VerbilligteSonderfertigkeit >();

        verbindungen = 0;

        voraussetzungen = new ArrayList< Voraussetzung >();

        vorteile = new ArrayList< VorNachteilBonus >();

        zauberboni = new ArrayList< Zauberbonus >();
    }

    /**
     * Berechnet, anhand der Boni, die GP-Kosten fuer den Baustein.
     *
     * @return die GP-Kosten fuer den Baustein.
     */
    public double berechneGpKosten() {

        double kosten = 0.0;

        for ( final EigenschaftsModifikation mod :
            eigenschaftsModifikationen.values() ) {

            kosten += mod.getGpKosten();
        }

        if ( hatMindeststufe( null, Eigenschaft.SO ) ) {

            final int stufe = getMindeststufe( Eigenschaft.SO );

            kosten += ( 0.5 * stufe );
        }

        for ( final BasiswertModifikation mod :
            basiswertModifikationen.values() ) {

            kosten += mod.getGpKosten();
        }

        for ( final VorNachteilBonus bonus : vorteile ) {

            kosten += bonus.getGpKosten();
        }

        kosten += Math.round( ( verbindungen / 3.0F ) * 0.3F );

        for ( final VorNachteilBonus bonus : nachteile ) {

            kosten += bonus.getGpKosten();
        }

        for ( final SonderfertigkeitBonus bonus : sonderfertigkeiten ) {

            kosten += bonus.getGpKosten();
        }

        for ( final VerbilligteSonderfertigkeit verbilligte :
            verbilligteSonderfertigkeiten ) {

            kosten += verbilligte.getGpKosten();
        }

        for ( final VerbilligteVariante verbilligte : verbilligteVariante ) {

            kosten += verbilligte.getGpKosten();
        }

        for ( final List< Talentbonus> boni : talentboni.values() ) {

            for ( final Talentbonus bonus : boni ) {

                kosten += bonus.getGpKosten();
            }
        }

        double kostenZauber = 0.0;

        for ( final Zauberbonus bonus : zauberboni ) {

            kostenZauber += bonus.getGpKosten();
        }

        final boolean paketVollzauberer = enthaeltVollzauberer();

        if ( paketVollzauberer ) {

            // Das Pakte Vollzauberer bringt Repraesentation
            // und Ritualkenntnis kostenfrei.

            if ( enthaelt( Sonderfertigkeit.REPRAESENTATION ) ) {

                kosten -= ( Sonderfertigkeit.REPRAESENTATION.getApKosten()
                            / 100.0 );
            }

            int ritualkenntnis = getBonus( Talent.RITUALKENNTNIS );

            if ( 0 <= ritualkenntnis ) {

                // eine Extrapunkt fuer die Aktivierung
                ++ritualkenntnis;

                if ( 4 < ritualkenntnis ) {

                    // nur 3 Punkte plus Aktivierung sind kostenfrei.
                    ritualkenntnis = 4;
                }

                kosten -= ( ( Talent.RITUALKENNTNIS.getSteigerungsfaktor()
                        * ritualkenntnis ) / 10.0 );

            }

            // Das Paket Vollzauberer bringt 12GP fuer Zauber
            kostenZauber -= 12;

            kostenZauber = Math.max( kostenZauber, 0 );
        }

        kosten += kostenZauber;

        kosten -= gpKostenAbzug;

        return kosten;
    }

    private int getBonus( final Talent ritualkenntnis ) {

        int bonus = -1;

        final List< Talentbonus > liste =
            talentboni.get( Talentgruppe.RITUALKENNTNIS );

        if ( ( null != liste ) && ( 0 < liste.size() ) ) {

            final Talentbonus tb = liste.get( 0 );

            switch( tb.getTyp() ) {

            case FEST:
            case ENTDECKER:
                bonus = ((TalentbonusFest) tb).getBoni().get( 0 );
                break;

            case FREI:
                bonus = ((TalentbonusFrei) tb).getStufenBonus();
            }
        }

        return bonus;
    }

    boolean enthaelt( final Sonderfertigkeit sonderfertigkeit ) {

        for ( final SonderfertigkeitBonus bonus : sonderfertigkeiten ) {

            if ( bonus.enthaelt( sonderfertigkeit ) ) {

                return true;
            }
        }

        return false;
    }

    /**
     * @return {@code true}, wenn der Baustein den Vorteil Vollzauberer
     *         enthaelt, sonst {@code false}.
     */
    private boolean enthaeltVollzauberer() {

        final VorNachteilWert vollzauberer =
            new VorNachteilWert( VorNachteil.VOLLZAUBERER );

        for ( final VorNachteilBonus bonus : vorteile ) {

            if ( bonus.getWerte().contains( vollzauberer ) ) {

                return true;
            }
        }

        return false;
    }

    /**
     * Fuegt dem Baustein einen neuen Talentbonus hinzu.
     *
     * @param bonus den der Baustein vergibt.
     */
    private void einfuegen( final Talentbonus bonus ) {
        talentboni.get( bonus.getTalentgruppe() ).add( bonus );
    }

    /**
     * Fuegt dem Baustein eine neue Voraussetzung hinzu.
     *
     * @param voraussetzung fuer den Baustein.
     */
    private void einfuegen( final Voraussetzung voraussetzung ) {
        voraussetzungen.add( voraussetzung );
    }

    /**
     * Fuegt dem Baustein eine neue Voraussetzung hinzu.
     *
     * @param voraussetzung fuer den Baustein.
     */
    private void einfuegen(
            final VoraussetzungEigenschaft voraussetzung ) {

        voraussetzungen.add( voraussetzung );

        // Es reicht, wenn die Voraussetzung fuer eine der Eigenschaften
        // erfuellt ist, daher wird hier nur die Eigenschaft an Position 0
        // abgegriffen.
        final Eigenschaft eigenschaft = voraussetzung.getEigenschaften()[ 0 ];

        eigenschaftsVoraussetzungen.put( eigenschaft, voraussetzung );

    }

    abstract IXMLElement erzeugeXmlWurzel();

    /**
     * @return alle Eigenschaft-Modifikationen des Bausteins.
     */
    public final
    List< EigenschaftsModifikation > getEigenschaftsModifikationen() {

        return new ArrayList< EigenschaftsModifikation >(
                eigenschaftsModifikationen.values() );
    }

    /**
     * @return eine sortierte Liste der empfohlenen Nachteile.
     */
    public final List< VorNachteil > getEmpfohleneNachteile() {
        return new ArrayList< VorNachteil >( empfohleneNachteile );
    }

    /**
     * @return eine sortierte Liste aller empfohlenen Vor- und Nachteile.
     */
    public Collection< VorNachteil > getEmpfohleneVorNachteile() {
        return new ArrayList< VorNachteil >( empfohleneVorNachteile );
    }

    /**
     * @return eine sortierte Liste der empfohlenen Vorteile.
     */
    public final List< VorNachteil > getEmpfohleneVorteile() {
        return new ArrayList< VorNachteil >( empfohleneVorteile );
    }

    /**
     * @param held der den Baustein waehlen will.
     * @return die GP-Kosten des Bausteins.
     */
    public final int getGpKosten( final Held held ) {
        return kosten.getGpKosten( held );
    }

    /**
     * Liefert die Hoechststufe welche der Baustein in einer Eigenschaft
     * zu laesst.
     *
     * @see #hatHoechststufe(Eigenschaft)
     * @param eigenschaft deren Hoechststufe gefragt ist.
     * @return die Hoechststufe die der Baustein in der Eigenschaft vor gibt.
     *      Wenn der Baustein in der Eigenschaft keine Hoechststufe vor gibt,
     *      dann ist der gelieferte Wert unbestimmt oder es wird eine Exception
     *      geworfen.
     * @throws NullPointerException in einigen Faellen, wenn es keine
     *      Hoechststufe fuer die Eigenschaft gibt.
     */
    public final int getHoechststufe( final Eigenschaft eigenschaft ) {
        return eigenschaftsVoraussetzungen.get( eigenschaft ).getHoechststufe();
    }

    /**
     * @return Das Objekt, das die Kosten berechnet.
     */
    public final Kosten getKostenObjekt() {
        return kosten;
    }

    /**
     * @see #hatLeittalente()
     * @return eine Liste aller Leittalente, oder eine leere Liste, wenn der
     *      Baustein keine Leittalente vergibt. Talent die durch einen
     *      {@link TalentbonusFest} zum Leittalent werden muessen in dieser
     *      Collection nicht vertreten sein. Die Collection enthaelt vor allem
     *      Talente die Leittalente sind, aber nicht durch den Baustein
     *      aktiviert werden. Sieh hierzu Regelwerk, Seite 82, Kasten.
     */
    public final List< Talent > getLeittalente() {
        return new ArrayList< Talent >( leittalente );
    }

    /**
     * @see #hatMindeststufe(Held, Eigenschaft)
     * @param eigenschaft deren Mindeststufe gefragt ist.
     * @return die Mindeststufe die der Baustein in der Eigenschaft vorgibt.
     *      Wenn keine Mindeststufe vorgegeben wird ist der gelieferte Wert
     *      unbestimmt oder es wird eine Exception geworfen.
     * @throws NullPointerException in einigen Faellen wenn es keine
     *      Mindeststufe fuer die Eigenschaft gibt.
     */
    public final int getMindeststufe( final Eigenschaft eigenschaft ) {
        return eigenschaftsVoraussetzungen.get( eigenschaft ).getMindeststufe();
    }

    /**
     * @see #hatModifikation(Basiswert)
     * @param basiswert dessen Modifikation gesucht ist.
     * @return die Modifikation die der Baustein auf den Basiswert vergibt.
     * @throws NullPointerException wenn es fuer den Basiswert keine
     *      Modifikation gibt.
     */
    public final int getModifikation( final Basiswert basiswert ) {
        return basiswertModifikationen.get( basiswert ).getModifikation();
    }

    /**
     * @see #hatModifikation(Eigenschaft)
     * @param eigenschaft deren Modifikation gesucht ist.
     * @return die Modifikation die der Baustein fuer die Eigenschaft vergibt.
     * @throws NullPointerException wenn es fuer die Eigenschaft keine
     *      Modifikation gibt.
     */
    public final int getModifikation( final Eigenschaft eigenschaft ) {
        return eigenschaftsModifikationen.get( eigenschaft ).getModifikation();
    }

    /**
     * @return die Eigenschafts- und Basiswert-Modifikationen des Bausteins.
     */
    public final List< Object > getModifikationen() {

        final List< Object > modifikationen;

        modifikationen = new ArrayList< Object >();

        modifikationen.addAll( basiswertModifikationen.values() );

        modifikationen.addAll( eigenschaftsModifikationen.values() );

        return modifikationen;
    }

    /**
     * @return eine Liste der Nachteile, die man automatisch durch den Baustein
     *         erhaelt.
     */
    public final List< VorNachteilBonus > getNachteile() {
        return new ArrayList< VorNachteilBonus >( nachteile );
    }

    /**
     * @return eine Liste aller Boni auf Sonderfertigkeiten, die ein Held,
     *      ueber diesen Baustein erhalten kann.
     */
    public final List< SonderfertigkeitBonus > getSonderfertigkeiten() {
        return new ArrayList< SonderfertigkeitBonus >( sonderfertigkeiten );
    }

    /**
     * @return eine Liste aller Talentboni.
     */
    public final Vector< Talentbonus > getTalentboni() {

        final Vector< Talentbonus > boni = new Vector< Talentbonus >();

        for ( final List< Talentbonus > liste : talentboni.values() ) {

            boni.addAll( liste );
        }

        return boni;
    }

    /**
     * Liefert die Talenboni einer bestimmten Talentgruppe.
     *
     * @param talentgruppe deren Boni gefragt sind. Die
     *      {@link Talentgruppe#GABEN} ist nicht zugelassen.
     * @return eine Liste der Talentboni der Gruppe.
     */
    public final List< Talentbonus > getTalentboni(
            final Talentgruppe talentgruppe ) {

        return new ArrayList< Talentbonus >( talentboni.get( talentgruppe ) );
    }

    /**
     * @return eine Liste der ungeeigneten Nachteile.
     */
    public final List< UngeeigneterVorNachteil > getUngeeigneteNachteile() {
        return new ArrayList< UngeeigneterVorNachteil >( ungeeigneteNachteile );
    }

    /**
     * @return eine Liste der ungeeigneten Vor- und Nachteile.
     */
    public final List< UngeeigneterVorNachteil > getUngeeigneteVorNachteile() {
        return new ArrayList< UngeeigneterVorNachteil >(
                ungeeigneteVorNachteile.values() );
    }

    /**
     * @return eine Liste der ungeeigneten Vorteile.
     */
    public final List< UngeeigneterVorNachteil > getUngeeigneteVorteile() {
        return new ArrayList< UngeeigneterVorNachteil >( ungeeigneteVorteile );
    }

    /**
     * @return die Sonderfertigkeiten die ein Held durch den Baustein als
     *      verbilligt erhaelt.
     */
    public final
    List< VerbilligteSonderfertigkeit > getVerbilligteSonderfertigkeiten() {

        return new ArrayList< VerbilligteSonderfertigkeit >(
                verbilligteSonderfertigkeiten );
    }

    /**
     * @return die Sonderfertigkeiten von denen nur eine Variante verbilligt
     *         wird.
     */
    public final  List< VerbilligteVariante > getVerbilligteVarianten() {
        return new ArrayList< VerbilligteVariante >(
                verbilligteVariante );
    }

    /**
     * @return die Punkte die auf den Vorteil Verbindungen verteilt werden
     *         duerfen. Siehe Profession Streuner.
     */
    public final int getVerbindungen() {
        return verbindungen;
    }

    /**
     * @return eine Liste aller Voraussetzungen des Bausteins.
     */
    public final List< Voraussetzung > getVoraussetzungen() {
        return new ArrayList< Voraussetzung >( voraussetzungen );
    }

    /**
     * @return eine Liste der Vorteile, die man automatisch durch den Baustein
     *         erhaelt.
     */
    public final List< VorNachteilBonus > getVorteile() {
        return new ArrayList< VorNachteilBonus >( vorteile );
    }

    /**
     * @return die Zauber, die man durch den Baustein erhaelt.
     */
    public final List< Zauberbonus > getZauberboni() {
        return new ArrayList< Zauberbonus >( zauberboni );
    }
    
    /**
     * @param vorNachteil
     *            fuer den geprueft werden soll ob er empfohlen ist.
     * @return {@code true} wenn der Vor-/Nachteil empfohlen ist, sonst
     *         {@code false}.
     */
    public final boolean hatEmpfohlenen( final VorNachteil vorNachteil ) {
        return empfohleneVorNachteile.contains( vorNachteil );
    }

    /**
     * Prueft ob der Baustein fuer eine Eigenschaft eine Hoechststufe vorgibt,
     * die ein Held nicht ueberschreiten darf, wenn er diesen Baustein waehlen
     * will.
     *
     * @param eigenschaft in der die Hoechststufe gefordert sein soll.
     * @return {@code true} wenn es eine Hoechststufe gibt, sonst
     *      {@code false}.
     */
    public final boolean hatHoechststufe( final Eigenschaft eigenschaft ) {

        if ( eigenschaftsVoraussetzungen.containsKey( eigenschaft ) ) {

            final VoraussetzungEigenschaft voraussetzung =
                eigenschaftsVoraussetzungen.get( eigenschaft );

            return voraussetzung.hatHoechststufe();

        } else {

            return false;

        }
    }

    /**
     * @return {@code true} wenn der Baustein Leittalente vergibt, sonst
     *      {@code false}.
     */
    public final boolean hatLeittalente() {
        return 0 < leittalente.size();
    }

    /**
     * Prueft ob der Baustein fuer einen Helden in einer Eigenschaft eine
     * Mindeststufe vorgibt die der Held mindestens erreichen muss.
     *
     * @param held fuer den die Mindeststufe sein soll.
     * @param eigenschaft in der die Mindesstufe gefordert sein soll.
     * @return {@code true} wenn es eine Mindesstufe gibt, sonst
     *      {@code false}.
     */
    public final boolean hatMindeststufe(
            final Held held,
            final Eigenschaft eigenschaft ) {

        if ( eigenschaftsVoraussetzungen.containsKey( eigenschaft ) ) {

            final VoraussetzungEigenschaft voraussetzung =
                    eigenschaftsVoraussetzungen.get( eigenschaft );

            return voraussetzung.hatMindeststufe( held );

        } else {

            return false;

        }
    }

    /**
     * Prueft ob der Baustein eine Modifikation fuer einen besimmten Basiswert
     * vergibt.
     *
     * @param basiswert der ueberprueft werden soll.
     * @return {@code true} wenn der Baustein eine Modifikation fuer den
     *      Basiswert vergibt, sonst {@code false}.
     */
    public final boolean hatModifikation( final Basiswert basiswert ) {
        return basiswertModifikationen.containsKey( basiswert );
    }

    /**
     * Prueft ob der Baustein eine Modifikation fuer eine besimmte Eigenschaft
     * vergibt.
     *
     * @param eigenschaft die ueberprueft werden soll.
     * @return {@code true} wenn der Baustein eine Modifikation fuer die
     *      Eigenschaft vergibt, sonst {@code false}.
     */
    public final boolean hatModifikation( final Eigenschaft eigenschaft ) {
        return eigenschaftsModifikationen.containsKey( eigenschaft );
    }

    /**
     * @return {@code true}, wenn der Baustein Talentboni hat, sonst
     *         {@code false}.
     */
    private boolean hatTalentbonus() {

        for ( final List< Talentbonus > boni : talentboni.values() ) {

            if ( 0 != boni.size() ) {

                return true;
            }
        }

        return false;
    }

    /**
     * @param held
     *            fuer den geprueft werden soll ob der Vor-/Nachteil ungeeignet
     *            ist.
     * @param vorNachteil
     *            fuer den geprueft werden soll ob er ungeeignet ist.
     * @return {@code true} wenn der Vor-/Nachteil ungeeignet ist, sonst
     *         {@code false}.
     */
    public final boolean hatUngeeigneten(
            final Held held,
            final VorNachteil vorNachteil ) {

        return ungeeigneteVorNachteile.containsKey( vorNachteil )
            && ungeeigneteVorNachteile.get( vorNachteil )
                                                .istUngeeignetFuer( held );
    }

    /**
     * Fuellt die Map {@link #talentboni} mit einer Liste fuer jede
     * Talentgruppe.
     */
    private void initTalentBoni() {

        for ( final Talentgruppe talentgruppe :
            Talentgruppe.getTalentgruppenAlle() ) {

            talentboni.put( talentgruppe, new ArrayList< Talentbonus >() );
        }
    }

    /**
     * Prueft ob ein Held alle Voraussetzungen erfuellt um diesen Baustein
     * waehlen zu koennen.
     *
     * @param held der ueberprueft werden soll.
     * @return eine Beschreibung evtl. bestehender Konflikte.
     */
    public final KonfliktBeschreibung pruefen( final Held held ) {

        final KonfliktBeschreibung beschreibung = new KonfliktBeschreibung();

        for ( final Voraussetzung voraussetzung : voraussetzungen ) {

            beschreibung.einfuegen( voraussetzung.pruefen( held ) );
        }

        return beschreibung;
    }

    /**
     * Bestimmt die Modifikationen auf die Basiswerte.
     *
     * @param modifikationen Die Modifikationen auf die Basiswerte.
     */
    public final void setBasiswertModifikationen(
            final List< BasiswertModifikation > modifikationen ) {

        basiswertModifikationen.clear();

        for ( final BasiswertModifikation mod : modifikationen ) {

            basiswertModifikationen.put( mod.getBasiswert(), mod );

        }
    }

    /**
     * Bestimmt die Modifikationen die auf die Eigenschaften vergeben werden.
     *
     * @param modifikationen die auf die Eigenschaften vergeben werden.
     */
    public final void setEigenschaftModifikationen(
            final List< EigenschaftsModifikation > modifikationen ) {

        eigenschaftsModifikationen.clear();

        for ( final EigenschaftsModifikation mod : modifikationen ) {

            eigenschaftsModifikationen.put( mod.getEigenschaft(), mod );

        }
    }

    /**
     * Legt die empfohlenen Vor-/Nachteile fest.
     *
     * @param vorNachteile
     *            die als empfohlen gelten sollen.
     */
    public final void setEmpfohlene( final Collection< VorNachteil> vorNachteile ) {

        empfohleneNachteile.clear();
        empfohleneVorNachteile.clear();
        empfohleneVorteile.clear();

        empfohleneVorNachteile.addAll( vorNachteile );
        
        for ( final VorNachteil vorNachteil : vorNachteile ) {

            if ( vorNachteil.istVorteil() ) {

                empfohleneVorteile.add( vorNachteil );

            } else {

                empfohleneNachteile.add( vorNachteil );

            }
        }

        Collections.sort( empfohleneNachteile, VorNachteil.NAME_COMPARATOR );
        Collections.sort( empfohleneVorteile,  VorNachteil.NAME_COMPARATOR );
    }

    /**
     * @param dieKosten des Heldenbausteins.
     */
    public final void setKosten( final Kosten dieKosten ) {
        kosten = dieKosten;
    }

    /**
     * Ueber diese Methode koennen Talente als Leittalente markiert werden.
     * Das ist nur fuer Talente noetig die nicht durch einen Talent-Bonus
     * als Leittalent markiert werden.
     *
     * @param dieLeittalente des Heldenbausteins.
     */
    public final void setLeittalente(
            final Collection< Talent > dieLeittalente ) {

        leittalente.clear();

        leittalente.addAll( dieLeittalente );
    }

    /**
     * @param modifikation die der Held auf einen Basiswert erhalten soll.
     */
    public final void setModifikation( final BasiswertModifikation modifikation ) {
       basiswertModifikationen.put( modifikation.getBasiswert(), modifikation );
    }

    /**
     * @param dieNachteile die ein Held automatisch durch den Baustein erhaelt.
     */
    public final void setNachteile( final List< VorNachteilBonus > dieNachteile ) {

        nachteile.clear();

        nachteile.addAll( dieNachteile );
    }

    /**
     * @param dieSonderfertigkeiten die man durch den Heldenbaustein erhaelt.
     */
    public final void setSonderfertigkeiten(
            final List< SonderfertigkeitBonus > dieSonderfertigkeiten ) {

        sonderfertigkeiten.clear();

        sonderfertigkeiten.addAll( dieSonderfertigkeiten );
    }

    /**
     * @param boni die Talent-Boni die ein Held durch den Baustein erhaelt.
     */
    public final void setTalente( final List< Talentbonus > boni ) {

        talentboni.clear();

        initTalentBoni();

        for ( final Talentbonus bonus : boni ) {

            einfuegen( bonus );
        }
    }

    /**
     * Legt die ungeeigneten Vor-/Nachteile fest.
     *
     * @param vorNachteile
     *            die als ungeeignet gelten sollen.
     */
    public final void setUngeeignete(
            final Collection< UngeeigneterVorNachteil > vorNachteile ) {

        ungeeigneteNachteile.clear();
        ungeeigneteVorNachteile.clear();
        ungeeigneteVorteile.clear();

        for ( final UngeeigneterVorNachteil vorNachteil : vorNachteile ) {

            ungeeigneteVorNachteile.put(
                    vorNachteil.getVorNachteil(), vorNachteil );

            if ( vorNachteil.getVorNachteil().istVorteil() ) {

                ungeeigneteVorteile.add( vorNachteil );

            } else {

                ungeeigneteNachteile.add( vorNachteil );

            }
        }

        Collections.sort( ungeeigneteNachteile );
        Collections.sort( ungeeigneteVorteile );
    }

    /**
     * @param verbilligte Die Sonderfertigkeiten die durch den Heldenbaustein
     *      verbilligt werden.
     */
    public final void setVerbilligte(
            final List< VerbilligteSonderfertigkeit > verbilligte ) {

        verbilligteSonderfertigkeiten.clear();

        verbilligteSonderfertigkeiten.addAll( verbilligte );
    }

    /**
     * @param verbilligte
     *            Die Sonderfertigkeiten von denen Varianten verbilligt sind.
     */
    public final void setVerbilligteVariante(
            final List< VerbilligteVariante > verbilligte ) {

        verbilligteVariante.clear();

        verbilligteVariante.addAll( verbilligte );
    }

    /**
     * @param punkte
     *            die auf den Vorteil Verbindungen verteilt werden duerfen. Siehe
     *            Profession Streuner.
     */
    public final void setVerbindungen( final int punkte ) {
        verbindungen = punkte;
    }

    /**
     * Aendert die Voraussetzungen des Baustein auf die in der Liste
     * uebergebenen.
     *
     * @param neueVoraussetzungen fuer den Heldenbaustein.
     */
    public final void setVoraussetzungen(
            final List< Voraussetzung > neueVoraussetzungen ) {

        voraussetzungen.clear();
        eigenschaftsVoraussetzungen.clear();

        for ( final Voraussetzung voraussetzung : neueVoraussetzungen ) {

            if ( voraussetzung instanceof VoraussetzungEigenschaft ) {

                einfuegen( (VoraussetzungEigenschaft) voraussetzung );

            } else {

                einfuegen( voraussetzung );

            }
        }
    }

    /**
     * @param dieVorteile die ein Held automatisch durch den Baustein erhaelt.
     */
    public final void setVorteile( final List< VorNachteilBonus > dieVorteile ) {

        vorteile.clear();

        vorteile.addAll( dieVorteile );
    }

    /**
     * @param dieZauber die ein Held durch den Baustein erhaelt.
     */
    public final void setZauberboni( final List< Zauberbonus > dieZauber ) {

        zauberboni.clear();

        zauberboni.addAll( dieZauber );
    }

    /** @return den Namen des Bausteins. */
    @Override
    public final String toString() {
        return getName();
    }

    /**
     * @return Die XML-Repraesentation des Bausteins.
     */
    public IXMLElement toXml() {

        final IXMLElement wurzelElement = erzeugeXmlWurzel();

        // Id
        // **
        final IXMLElement idElement = new XMLElement( TAG_ID );

        idElement.setContent( getId() );

        wurzelElement.addChild( idElement );

        // Kosten
        // ******
        wurzelElement.addChild( kosten.toXml() );

        // Voraussetzungen
        // ***************
        if ( 0 != voraussetzungen.size() ) {

            final IXMLElement voraussetzungenElement =
                new XMLElement( TAG_VORAUSSETZUNGEN );

            for ( final Voraussetzung voraussetzung : voraussetzungen ) {

                voraussetzungenElement.addChild( voraussetzung.toXml() );
            }

            wurzelElement.addChild( voraussetzungenElement );
        }

        // Modifikationen
        // **************
        if ( 0 != eigenschaftsModifikationen.size()
          || 0 != basiswertModifikationen.size() ) {

            final IXMLElement modifikationenElement =
                new XMLElement( TAG_MODIFIKATIONEN );

            for ( final EigenschaftsModifikation mod :
                            eigenschaftsModifikationen.values() ) {

                modifikationenElement.addChild( mod.toXml() );
            }

            for ( final BasiswertModifikation mod :
                            basiswertModifikationen.values() ) {

                modifikationenElement.addChild( mod.toXml() );
            }

            wurzelElement.addChild( modifikationenElement );
        }

        // Vorteile
        // ********
        if ( 0 != vorteile.size() ) {

            final IXMLElement vorteileElement =
                new XMLElement( TAG_VORTEIL_BONI );

            for ( final VorNachteilBonus bonus : vorteile ) {

                vorteileElement.addChild( bonus.toXml() );
            }

            wurzelElement.addChild( vorteileElement );
        }

        // Verbindungen
        // ************
        if ( 0 != verbindungen ) {

            final IXMLElement verbindungenElement =
                new XMLElement( TAG_VERBINDUNGEN );

            verbindungenElement.setContent( String.valueOf( verbindungen ) );

            wurzelElement.addChild( verbindungenElement );
        }

        // Nachteile
        // *********
        if ( 0 != nachteile.size() ) {

            final IXMLElement vorteileElement =
                new XMLElement( TAG_NACHTEIL_BONI );

            for ( final VorNachteilBonus bonus : nachteile ) {

                vorteileElement.addChild( bonus.toXml() );
            }

            wurzelElement.addChild( vorteileElement );
        }

        // empfohlene Vor-/Nachteile
        // *************************
        final IXMLElement empfohleneElement =
            FactoryEmpfohleneVorNachteileOut.erzeugen( this );

        if ( null != empfohleneElement ) {

            wurzelElement.addChild( empfohleneElement );
        }

        // ungeeignete Vor-/Nachteile
        // **************************
        if ( 0 != ungeeigneteVorNachteile.size() ) {

            final IXMLElement ungeeigneteElement =
                new XMLElement( TAG_UNGEEIGNETE_VOR_NACH );

            for ( final UngeeigneterVorNachteil ungeeigneter :
                                            ungeeigneteVorNachteile.values() ) {

                ungeeigneteElement.addChild( ungeeigneter.toXml() );
            }

            wurzelElement.addChild( ungeeigneteElement );
        }

        // Talente
        // *******
        if ( hatTalentbonus() ) {

            final IXMLElement talenteElement = new XMLElement( TAG_TALENT_BONI);

            for ( final List< Talentbonus > liste : talentboni.values() ) {

                for ( final Talentbonus bonus : liste ) {

                    talenteElement.addChild( bonus.toXml() );
                }
            }

            wurzelElement.addChild( talenteElement );
        }

        // Leittalente
        // ***********
        if ( 0 != leittalente.size() ) {

            final IXMLElement leittalenteElement =
                new XMLElement( TAG_LEITTALENTE );

            for ( final Talent leittalent : leittalente ) {

                final IXMLElement talentElement = new XMLElement( TAG_TALENT );

                talentElement.setContent( leittalent.getId() );

                leittalenteElement.addChild( talentElement );
            }

            wurzelElement.addChild( leittalenteElement );
        }

        // Zauber
        // ******
        if ( 0 != zauberboni.size() ) {

            final IXMLElement zauberboniElement =
                new XMLElement( TAG_ZAUBER_BONI );

            for ( final Zauberbonus bonus : zauberboni ) {

                zauberboniElement.addChild( bonus.toXml() );
            }

            wurzelElement.addChild( zauberboniElement );
        }

        // Sonderfertigkeiten
        // ******************
        if ( 0 != sonderfertigkeiten.size() ) {

            final IXMLElement sonderfertigkeitenElement =
                new XMLElement( TAG_SONDERFERTIGKEIT_BONI );

            for ( final SonderfertigkeitBonus bonus : sonderfertigkeiten ) {

                sonderfertigkeitenElement.addChild( bonus.toXml() );
            }

            wurzelElement.addChild( sonderfertigkeitenElement );
        }

        // verbilligte Sonderfertigkeiten
        // ******************************
        if ( 0 != verbilligteSonderfertigkeiten.size() ) {

            final IXMLElement verbSfenElement =
                new XMLElement( TAG_VERB_SONDERFERTIGKEITEN );

            for ( final VerbilligteSonderfertigkeit verbSf :
                                    verbilligteSonderfertigkeiten ) {

                verbSfenElement.addChild( verbSf.toXml() );
            }

            wurzelElement.addChild( verbSfenElement );
        }

        // verbilligte Varianten von Sonderfertigkeiten
        // ********************************************
        if ( 0 != verbilligteVariante.size() ) {

            final IXMLElement verbVariantenElement =
                new XMLElement( TAG_VERBILLIGTE_VARIANTEN );

            for ( final VerbilligteVariante verbilligt :
                                    verbilligteVariante ) {

                verbVariantenElement.addChild( verbilligt.toXml() );
            }

            wurzelElement.addChild( verbVariantenElement );
        }

        return wurzelElement;
    }
}

