/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.bausteine;

import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_AUGENFARBEN;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_HAARFARBEN;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_KULTUR;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_KULTUREN;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_NAME;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_RASSE;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;

import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;
import de.bernhardjung.dsaprogramm.basis.Augenfarbe;
import de.bernhardjung.dsaprogramm.basis.Haarfarbe;
import de.bernhardjung.dsaprogramm.basis.Wuerfel;
import de.bernhardjung.dsaprogramm.bausteine.util.AlterFormel;
import de.bernhardjung.dsaprogramm.bausteine.util.GewichtFormel;
import de.bernhardjung.dsaprogramm.bausteine.util.GroesseFormel;
import de.bernhardjung.dsaprogramm.bausteine.util.Wuerfeltabelle;

public class Rasse
extends Heldenbaustein {

    /**
     * Eine Rasse die keine Kosten verursacht und keinerlei Boni vergibt.
     * Sie kann als Platzhalter dienen wenn fuer einen Helden noch keine
     * Rasse gewaehlt wurde.
     */
    public static final Rasse LEERE_RASSE = new Rasse( "Keine Rasse" );

    /** Die Formel zur Bestimmung des Alters eines Helden. */
    private AlterFormel alter;

    /** Die Wuerfeltabelle fuer die Augenfarben. */
    private Wuerfeltabelle augenfarben;

    /** Die Formel zur Berechnung des Gewichts eines Helden. */
    private GewichtFormel gewichtFormel;

    /** Die Formel zu Bestimmung der Koerpergroesse eines Helden. */
    private GroesseFormel groesseFormel;

    /** Die Wuerfeltabelle fuer die Haarfarben. */
    private Wuerfeltabelle haarfarben;

    /**
     * Die mit der Rasse verbunden Kulturen.
     * Diese Kulturen duerfen laut Regeln gewaehlt werden wenn man diese Rasse
     * waehlt.
     */
    private final Collection< Kultur > kulturen;

    /**
     * Erzeugt eine neue Rasse und initialisert alle Werte.
     *
     * @param id der Rasse.
     */
    public Rasse( final String id ) {

        super( id, 8.5 );

        alter = new AlterFormel( 15, Wuerfel.W3 );

        augenfarben = new Wuerfeltabelle();
        augenfarben.eintragen( 1, 20, Augenfarbe.BRAUN.getId() );

        gewichtFormel = new GewichtFormel( 100 );

        groesseFormel = new GroesseFormel( 160, 2, 0 );

        haarfarben = new Wuerfeltabelle();
        haarfarben.eintragen( 1, 20, Haarfarbe.BRAUN.getId() );

        kulturen = new ArrayList< Kultur >();
    }

    @Override
    IXMLElement erzeugeXmlWurzel() {
        return new XMLElement( TAG_RASSE );
    }

    /**
     * @return Die Formel zum Auswuerfeln des Alters.
     */
    public final AlterFormel getAlter() {
        return alter;
    }

    /**
     * @return Die Wuerfeltabelle fuer die Augenfarben.
     */
    public final Wuerfeltabelle getAugenfarben() {
        return augenfarben;
    }

    /**
     * @return die Formel zur Berechnung des Gewichts eines Helden.
     */
    public final GewichtFormel getGewicht() {
        return gewichtFormel;
    }

    /**
     * @return die Formel zur Bestimmung der Koerpergroesse eines Helden.
     */
    public final GroesseFormel getGroesse() {
        return groesseFormel;
    }

    /**
     * @return Die Wuerfeltabelle fuer die Haarfarben.
     */
    public final Wuerfeltabelle getHaarfarben() {
        return haarfarben;
    }

    /**
     * @return die Kulturen die gewaehlt werden duerfen wenn man diese Rasse
     *      gewaehlt hat.
     */
    public final Collection< Kultur > getKulturen() {
        return new HashSet< Kultur >( kulturen );
    }

    /**
     * @param dasAlter die Formel zum Auswuerfeln des Alters.
     */
    public void setAlter( final AlterFormel dasAlter ) {
        alter = dasAlter;
    }

    /**
     * @param wuerfeltabelle fuer die Augenfarben.
     */
    public void setAugenfarben( final Wuerfeltabelle wuerfeltabelle ) {

        augenfarben = wuerfeltabelle;
    }

    /**
     * Legt die Formel fuer die Berechnung des Gewichts eines Helden fest.
     * @param formel fuer die Berechnung des Gewichts.
     */
    public void setGewicht( final GewichtFormel formel ) {
        gewichtFormel = formel;
    }

    /**
     * Legt die Formel zur Bestimmung der Koerpergroesse eines Helden fest.
     *
     * @param formel zur Bestimmung der Koerpergroesse.
     */
    public void setGroesse( final GroesseFormel formel ) {
        groesseFormel = formel;
    }

    /**
     * @param wuerfeltabelle fuer die Haarfarbe.
     */
    public void setHaarfarben( final Wuerfeltabelle wuerfeltabelle ) {

        haarfarben = wuerfeltabelle;
    }

    /**
     * @param dieKulturen die moeglichen Kulturen der Rasse.
     */
    public void setKulturen( final Collection< Kultur > dieKulturen ) {

        kulturen.clear();

        kulturen.addAll( dieKulturen );
    }

    @Override
    public IXMLElement toXml() {

        final IXMLElement wurzelElement = super.toXml();

        // Der Name
        // ********
        if ( null != name ) {

            final IXMLElement nameElement = new XMLElement( TAG_NAME );

            nameElement.setContent( getName() );

            wurzelElement.addChild( nameElement );
        }

        // Alter
        // *****
        wurzelElement.addChild( alter.toXml() );

        // Haarfarben
        // **********
        final IXMLElement haarfarbenElement = haarfarben.toXml();

        haarfarbenElement.setName( TAG_HAARFARBEN );

        wurzelElement.addChild( haarfarbenElement );

        // Augenfarben
        // ***********
        final IXMLElement augenfarbenElement = augenfarben.toXml();

        augenfarbenElement.setName( TAG_AUGENFARBEN );

        wurzelElement.addChild( augenfarbenElement );

        // Groesse
        // *******
        wurzelElement.addChild( groesseFormel.toXml() );

        // Gewicht
        // *******
        wurzelElement.addChild( gewichtFormel.toXml() );

        // Kulturen
        // ********
        final IXMLElement kulturenElement = new XMLElement( TAG_KULTUREN );

        for ( final Kultur kultur : kulturen ) {

            final IXMLElement kulturElement = new XMLElement( TAG_KULTUR );

            kulturElement.setContent( kultur.getId() );

            kulturenElement.addChild( kulturElement );
        }

        wurzelElement.addChild( kulturenElement );

        return wurzelElement;
    }
}
