/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.bausteine.bonus;

import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.ATTR_BESCHREIBUNG;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.ATTR_SONDERFERTIGKEIT;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.ATTR_TALENT;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.ATTR_VARIANTE;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_VARIANTE;

import java.util.List;

import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;
import de.bernhardjung.dsaprogramm.basis.varianten.Variante;
import de.bernhardjung.dsaprogramm.i18n.I18n;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.Sonderfertigkeit;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.SonderfertigkeitWert;

public class VerbilligteVariante {

    /** Die Sonderfertigkeit die verbilligt wird. */
    private final Sonderfertigkeit sonderfertigkeit;

    /**
     * Die genaue Variante wird ja noch vom Benutzer festgelegt daher ist
     * es eigentlich nicht noetig explizit eine Anzugeben.
     * Im Fall der Ortskenntnis kann die Variante hier aber ein String mit
     * einem Erklaerungstext sein wie z.B. "eine ueblicherweise benutzte
     * Strecke".
     */
    private final Variante variante;

    /**
     * Der Text der Beschreibung oder die Id die, in der XML-Datei, fuer die
     * Variante verwendet wird.
     */
    private final String beschreibungId;

    /**
     * @param dieSonderfertigkeit
     *            die verbilligt wird.
     * @param dieVariante
     *            Eine Variante, die verwendet wird, um einen Wert zu 
     *            initialisieren.
     * @param dieBeschreibung
     *            der Verbilligung z.B. &quot;eine weiter Kulturkunde&quot; Wenn
     *            fuer {@code dieVariante} ein Text uebergeben wurde kann hier
     *            {@code null} uebergeben werden. Es kann hier auch eine Id,
     *            fuer einen Text aus der Lokalisiserungsdatei, uebergeben
     *            werden.
     * @throws IllegalArgumentException
     *             wenn fuer {@code dieVariante} und
     *             {@code dieBeschreibung null} uebergeben wurde.
     */
    public VerbilligteVariante(
            final Sonderfertigkeit dieSonderfertigkeit,
            final Variante dieVariante,
            final String dieBeschreibung ) {

        if ( ( null == dieVariante ) && ( null == dieBeschreibung ) ) {
            throw new IllegalArgumentException(
                    "dieVariante und dieBeschreibung sind null. " +
                    "Eines der Argumente muss ungleich null sein." );
        }

        sonderfertigkeit = dieSonderfertigkeit;

        variante = dieVariante;
        
        beschreibungId = dieBeschreibung;
    }

    /**
     * @return die Id der Beschreibung, oder {@code null}, wenn es keine Id fuer
     *         eine Beschreibung gibt.
     */
    public final String getBeschreibungId() {
        return beschreibungId;
    }

    public double getGpKosten() {
        return sonderfertigkeit.getApKosten() / 250.0;
    }

    /**
     * @return Die Variante falls es eine gibt oder {@code null}.
     */
    public final Variante getVariante() {
        
        return variante;
    }
    
    /**
     * @return Die Sonderfertigkeit die verbilligt wird.
     */
    public final SonderfertigkeitWert getWert() {
        
        final Variante var;
        
        if ( null == variante && sonderfertigkeit.hatVarianten() ) {
            
            final List< Variante > varianten = 
                sonderfertigkeit.getVarianten(null);
            
            if ( 0 == varianten.size() ) {
                
                var = new Variante( 
                        I18n.i18n( "VerbilligteVariante.FreiWaehlbar" ) );
            
            } else {
                
                var = varianten.get( 0 );
            }
        
        } else {
            
            var = variante;
        }
        
        return new SonderfertigkeitWert( sonderfertigkeit, var );
    }

    /**
     * @return eine Beschreibung der Verbilligung wie &quot;eine weitere
     *         Kulturkunde&quot;
     */
    @Override
    public final String toString() {
        
        if ( null != beschreibungId ) {

            if ( I18n.hatLokalisierungstext( beschreibungId ) ) {

                return I18n.i18n( beschreibungId );

            } else {

                return beschreibungId;
            }

        } else {

            return new SonderfertigkeitWert(
                    sonderfertigkeit,
                    variante ).toString();
        }
    }

    /**
     * @return das XML-Element, das die verbilligte Variante repraesentiert.
     */
    public final IXMLElement toXml() {

        final IXMLElement element = new XMLElement( TAG_VARIANTE );

        element.setAttribute( ATTR_SONDERFERTIGKEIT, sonderfertigkeit.getId() );

        if ( sonderfertigkeit.hatTalent() ) {

            element.setAttribute( ATTR_TALENT,
                                  sonderfertigkeit.getTalent().getId() );
        }

        if ( null != beschreibungId ) {

            element.setAttribute( ATTR_BESCHREIBUNG, beschreibungId );
        }

        if ( null != variante ) {

            element.setAttribute( ATTR_VARIANTE, variante.getId() );
        }

        return element;
    }
}
