/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.bausteine.bonus;

import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.ATTR_VARIANTE_AENDERBAR;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_WERTE;

import java.util.ArrayList;
import java.util.List;

import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;
import de.bernhardjung.dsaprogramm.util.Strings;
import de.bernhardjung.dsaprogramm.vornachteile.VorNachteilWert;

public class VorNachteilBonus {

    /** Die Werte aus denen gewaehlt werden darf. */
    private final List< VorNachteilWertBonus > boni;

    /** Gibt an, ob die Variante der Werte geaendert werden darf. */
    private final boolean varianteAenderbar;

    /**
     * @param dieBoni
     *            von denen einer gewaehlt werden darf. Die Liste kann auch nur
     *            einen Wert enthalten.
     * @param istVarianteAenderbar
     *            {@code true}, wenn die Variante der Werte, bei der
     *            Erschaffung, vom Spieler geaendert werden darf, sonst
     *            {@code false}.
     */
    public VorNachteilBonus(
            final List< VorNachteilWertBonus > dieBoni,
            final boolean istVarianteAenderbar ) {

        boni = new ArrayList< VorNachteilWertBonus >( dieBoni );

        varianteAenderbar = istVarianteAenderbar;
    }

    /**
     * @return die GP-Kosten die fuer einen Baustein durch den Bonus entstehen.
     */
    public double getGpKosten() {

        double kosten = 0.0;

        for ( final VorNachteilWertBonus bonus : boni ) {

            final VorNachteilWert wert = bonus.getWert();

            kosten += wert.getGpKosten();
        }

        kosten /= boni.size();

        kosten *= 0.3;

        System.out.printf( "%1$s: %2$.1f\n", boni, kosten );

        return kosten;
    }

    /**
     * @return die Werte von denen einer ausgewaehlt werden soll.
     */
    public final List< VorNachteilWert > getWerte() {

        final List< VorNachteilWert > werte =
            new ArrayList< VorNachteilWert >( boni.size() );

        for ( final VorNachteilWertBonus bonus : boni ) {

            werte.add( bonus.getWert() );
        }

        return werte;
    }

    /** @return eine Aufzaehlung der waehlbaren Werte. */
    @Override
    public final String toString() {

        return Strings.erzeugeAufzaehlungMitOder( boni );
    }

    /**
     * @return Ein XML-Element, das den Bonus repraesentiert.
     */
    public final IXMLElement toXml() {

        final IXMLElement element = new XMLElement( TAG_WERTE );

        if ( varianteAenderbar ) {

            element.setAttribute( ATTR_VARIANTE_AENDERBAR, "true" );
        }

        for ( final VorNachteilWertBonus bonus : boni ) {

            element.addChild( bonus.toXml() );
        }

        return element;
    }
}
