/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.erschaffung;

import java.awt.event.ActionEvent;
import java.util.List;
import java.util.Observable;
import java.util.Observer;

import javax.swing.AbstractAction;

import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.bausteine.Kultur;
import de.bernhardjung.dsaprogramm.bausteine.Profession;
import de.bernhardjung.dsaprogramm.bausteine.Rasse;
import de.bernhardjung.dsaprogramm.bausteine.bonus.SonderfertigkeitBonus;
import de.bernhardjung.dsaprogramm.bausteine.bonus.Talentbonus;
import de.bernhardjung.dsaprogramm.bausteine.bonus.TalentbonusFest;
import de.bernhardjung.dsaprogramm.bausteine.bonus.TalentbonusFrei;
import de.bernhardjung.dsaprogramm.bausteine.bonus.VerbilligteSonderfertigkeit;
import de.bernhardjung.dsaprogramm.bausteine.bonus.VorNachteilBonus;
import de.bernhardjung.dsaprogramm.daten.Datenspeicher;
import de.bernhardjung.dsaprogramm.gui.Hauptpanel;
import de.bernhardjung.dsaprogramm.gui.PanelErschaffung;
import de.bernhardjung.dsaprogramm.gui.PanelGp;
import de.bernhardjung.dsaprogramm.gui.PanelRasKulProf;
import de.bernhardjung.dsaprogramm.gui.PanelSteigernErschaffung;
import de.bernhardjung.dsaprogramm.gui.bonus.*;
import de.bernhardjung.dsaprogramm.gui.steigern.PanelSteigern;
import de.bernhardjung.dsaprogramm.i18n.I18n;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.Sonderfertigkeit;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.SonderfertigkeitWert;
import de.bernhardjung.dsaprogramm.talente.Sprache;
import de.bernhardjung.dsaprogramm.talente.TalentWert;
import de.bernhardjung.dsaprogramm.util.Uebersetzbar;
import de.bernhardjung.dsaprogramm.util.Uebersetzer;
import de.bernhardjung.dsaprogramm.util.Util;
import de.bernhardjung.dsaprogramm.vornachteile.VorNachteilWert;
import de.bernhardjung.dsaprogramm.zauber.Zauber;

public class Erschaffung implements Uebersetzbar {

    /** Action mit der die Erschaffung abgeschlossen wird. */
    private final ActionBeenden actionBeenden;

    /**
     * Die Action die nach dem Auswaehlen der Leitzauber ausgefuehrt wird.
     */
    private final ActionLeitzauber actionLeitzauber;

    /**
     * Die Action mit der nacheinander die offenen Auswahlen fuer die
     * Sonderfertigkeiten angewendet werden.
     */
    private final 
    ActionSonderfertigkeitBoniOffen actionSonderfertigkeitBoniOffen;

    /**
     * Action mit der zum 2ten Schritt der Erschaffung gegangen wird, dem
     * anwenden der Talent-Boni.
     */
    private final ActionTalentBoni actionTalentBoni;

    /**
     * Action mit der nacheinander die offenen Talent-Boni aus Rasse und
     * Kultur ausgewertet werden.
     */
    private final ActionTalentBoniOffen actionTalentBoniOffen;

    /**
     * Action mit der nacheinander die offenen Talent-Boni die der Held durch
     * seine Profession erhalten hat angewendet werden.
     */
    private final 
    ActionTalentBoniOffenProfession actionTalentBoniOffenProfession;

    /**
     * Die Action die verwendet wird, wenn aus einer Liste eine verbilligte
     * Sonderfertigkeit gewaehlt werden muss.
     */
    private final 
    ActionVerbilligteSonderfertigkeit actionVerbilligteSonderfertigkeit;

    /**
     * Die Action mit der die offenen Boni auf die Vor-/Nachteile abgearbeitet
     * werden.
     */
    private final ActionVorNachteilBoniOffen actionVorNachteilBoniOffen;

    /** Action die verwendet wird um die Zweitsprache des Helden zu waehlen. */
    private final ActionZweitsprache actionZweitsprache;

    /** Das Hauptfenster der Anwendung. */
    private final Hauptpanel hauptpanel;

    /** Der Manager fuer die aktuell laufende Erschaffung. */
    private ErschaffungManager manager;

    /** Das Panel fuer die Heldenerschaffung. */
    private PanelErschaffung panelErschaffung;

    /**
     * Das Panel mit dem die zusaetzlichen Leitzauber fuer eien elfischen
     * Helden gewaehlt werden koennen.
     */
    private PanelLeitzauber panelLeitzauber;

    /**
     * Die Schnittstelle die es erlaubt eine Sonderfertigkeit fuer den Helden
     * zu waehlen.
     */
    private PanelAuswahl< SonderfertigkeitWert > panelSonderfertigkeitWert;

    /**
     * Die Schnittstelle die es erlaubt eine Sonderfertigkeit aus einer Liste
     * zu waehlen.
     */
    private PanelAuswahl< Sonderfertigkeit > panelSonderfertigkeit;

    /**
     * Die Benutzerschnittstelle mit der eine Auswahl fuer einen Talent-Bonus
     * getroffen werden kann. Bei einigen Talent-Boni kann der Benuter waehlen,
     * z.B. 'Schwerter oder Saebel +5'.
     */
    private PanelTalentBonus panelTalentBonus;

    /** Erlaubt es einen Vor-/Nachteil auszuwaehlen. */
    private PanelAuswahl< VorNachteilWert > panelVorNachteilWert;

    /** Ein Panel das es erlaubt eine Zweitsprache zu waehlen. */
    private PanelAuswahl< Sprache > panelZweitsprache;

    /**
     *
     * @param dasHauptfenster der Anwendung.
     */
    public Erschaffung( final Hauptpanel dasHauptpanel ) {

        hauptpanel = dasHauptpanel;

        actionBeenden = new ActionBeenden();

        actionLeitzauber = new ActionLeitzauber();

        actionSonderfertigkeitBoniOffen =
            new ActionSonderfertigkeitBoniOffen();

        actionTalentBoni = new ActionTalentBoni();

        actionTalentBoniOffen = new ActionTalentBoniOffen();

        actionTalentBoniOffenProfession =
            new ActionTalentBoniOffenProfession();

        actionVerbilligteSonderfertigkeit =
            new ActionVerbilligteSonderfertigkeit();

        actionVorNachteilBoniOffen = new ActionVorNachteilBoniOffen();
        
        actionZweitsprache = new ActionZweitsprache();
    }

    /**
     * Setzt das {@link #panelTalentBonus} mit dessen Hilfe die Auswahl fuer
     * den uebergebenen Talent-Bonus getroffen wird und zeigt den Bonus im
     * Panel an.
     *
     * @param bonus fuer den der Benutzer als naechstes eine Auswahl treffen
     *      muss.
     */
    private void anwendenTalentBonus( final Talentbonus bonus ) {

        switch( bonus.getTyp() ) {

        case FEST: {

            final PanelTalentBonusFest panel;
            panel = panelErschaffung.getPanelTalentBonusFest();

            panelTalentBonus = panel;

            panel.setBonus( manager.getHeld(), (TalentbonusFest) bonus );

            break;
        }
        case FREI: {

            final PanelTalentBonusFrei panel;
            panel = panelErschaffung.getPanelTalentBonusFrei();

            panelTalentBonus = panel;

            panel.setBonus( manager.getHeld(), (TalentbonusFrei) bonus );

            break;
        }
        default:
            throw new IllegalStateException( String.format(
                    "Die Methode Erschaffung#anwendenTalentBonus(TalentBonus)"+
                    " wurde mit einem TalentBonus aufgerufen der nicht vom " +
                    "Typ FEST oder FREI ist sondern vom Typ %1$s.",
                    bonus.getTyp() ) );
        }

    }
    
    public ErschaffungManager debugGetManager() {
        
        return manager;
    }
    
    public void neuUebersetzen() {
        
        Uebersetzer.neuUebersetzen(
                actionBeenden,
                actionLeitzauber,
                actionSonderfertigkeitBoniOffen,
                actionTalentBoni,
                actionTalentBoniOffen,
                actionTalentBoniOffenProfession,
                actionVerbilligteSonderfertigkeit,
                actionVorNachteilBoniOffen,
                actionZweitsprache
        );
    }
    
    /**
     * Der Teilschritt 1 bei der Erschaffung des Helden im Programm,
     * es werden Rasse, Kultur und Profession gewaehlt.
     */
    private void schritt_1_Bausteine() {

        manager.getHeld().addObserver( actionTalentBoni );

        panelErschaffung.setWeiterAction( actionTalentBoni );

        final PanelRasKulProf panelRasKulProf =
            panelErschaffung.getPanelRasKulProf();

        panelRasKulProf.setHeld( manager.getHeld() );

        panelErschaffung.anzeigen( panelRasKulProf );

    }

    /**
     * Teilschritt 2 bei der Heldenerschaffung im Programm.
     * Es werden die Alle-Boni die der Held durch Rasse, Kulutur
     * und Profession erhaelt angewendet.
     */
    private void schritt_2_BoniAnwenden() {

        panelErschaffung.getPanelRasKulProf().abmeldenVonHeld();

        // Muss hier ausgefuehrt werden, damit die KL eingestellt wird fuer
        // das Anzeigen des TaW der Zweit- und Muttersprache.
        manager.einstellenMindeststufe();

        manager.anwendenTalentBoniLeittalente();

        manager.anwendenSonderfertigkeiten();

        manager.anwendenVorNachteile();

        manager.anwendenZauber();

        schritt_3_1_Zweitsprache();
    }

    /**
     * Teilschritt 3.1 bei der Heldenerschaffung im Programm,
     * wenn noetig kann eine Auswahl fuer die Zweitsprache des Helden
     * getroffen werden.
     */
    private void schritt_3_1_Zweitsprache() {

        if ( null != panelZweitsprache ) {
            // Es wurde gerade ueber das panelZweitsprache die Zweitsprache
            // des Helden gewaehlt. Das gilt es jetzt auszuwerten.

            final Sprache zweitsprache;
            zweitsprache = panelZweitsprache.getGewaehltesElement();

            manager.setZweitsprache( zweitsprache );

            panelZweitsprache = null;
        }

        if ( manager.hatZweitsprachenAuswahl() ) {

            final List< Sprache > zweitsprachen = manager.getZweitsprachen();

            panelZweitsprache = panelErschaffung.getPanelSprache();

            panelZweitsprache.setElemente(
                    "Erschaffung.ZweitspracheWaehlen",
                    zweitsprachen,
                    Sprache.NAME_COMPARATOR );

            panelErschaffung.anzeigen( panelZweitsprache );

            panelErschaffung.setWeiterAction( actionZweitsprache );

        } else {

            schritt_3_2_TalentBoniOffen();
        }
    }

    /**
     * Teilschritt 3.2 bei der Heldenerschaffung,
     * die offenen Talentboni die der Held durch Rasse und Kultur erhalten hat
     * werden abgearbeitet.
     */
    private void schritt_3_2_TalentBoniOffen() {

        if ( null != panelTalentBonus ) {
                // Es wurde gerade ueber panelTalentBonus ein Talent-Bonus
            // verteilt. Die Auswahl gilt es jetzt auszuwerten.

            for ( final TalentWert wert : panelTalentBonus.getTalentBoni() ) {

                manager.einfuegenVorgegeben( wert );
            }

            panelTalentBonus = null;
        }

        if ( manager.hatTalentBonusOffen() ) {

            anwendenTalentBonus( manager.getTalentBonusOffen() );

            panelErschaffung.anzeigen( panelTalentBonus );

            panelErschaffung.setWeiterAction( actionTalentBoniOffen );

        } else {

            // Es ist kein offener Bonus aus Rasse oder Kultur mehr zu
            // bearbeiten. Auf zum naechsten Schritt.
            schritt_3_3_TalentBoniOffenProfession();

        }
    }

    /**
     * Teilschritt 2.3 der Heldenerschaffung, die offenen Talent-Boni die der
     * Held durch seine Profession erhalten hat sind abzuarbeiten.
     */
    private void schritt_3_3_TalentBoniOffenProfession() {

        if ( null != panelTalentBonus ) {
            // Es wurde gerade ein Talent-Bonus ausgewaehlt. Die Asuwahl gilt
            // es nun auszuwerten.
            for ( final TalentWert wert : panelTalentBonus.getTalentBoni() ) {

                manager.einfuegenProfession( wert );

            }

            panelTalentBonus = null;
        }

        if ( manager.hatTalentBonusOffenProfession() ) {

            anwendenTalentBonus( manager.getTalentBonusOffenProfession() );

            panelErschaffung.anzeigen( panelTalentBonus );

            panelErschaffung.setWeiterAction( actionTalentBoniOffenProfession );

        } else if ( manager.hatTalentBonusEntdecker() ) {

            final TalentbonusFest bonus = manager.getTalentBonusEntdecker();

            final PanelTalentBonusFestEntdecker2 panel;
            panel = panelErschaffung.getPanelTalentBonusEntdecker();

            panel.setBonus( manager.getHeld(), bonus );

            panelTalentBonus = panel;

            panelErschaffung.anzeigen( panelTalentBonus );

            panelErschaffung.setWeiterAction( actionTalentBoniOffenProfession );

        } else {

            // Es ist kein offener Bonus aus der Profession mehr zu
            // bearbeiten. Auf zum naechsten Schritt.
            schritt_3_1_SonderfertigkeitBoniOffen();
        }
    }

    /**
     * Teilschritt 3.0 der Heldenerschaffung.
     * Der Benutzer kann nacheinander die Sonderfertigkeit-Boni abarbeiten fuer
     * die eine Auswahl getroffen werden muss.
     */
    private void schritt_3_1_SonderfertigkeitBoniOffen() {

        if ( null != panelSonderfertigkeitWert ) {
            // Es wurde gerade eine Sonderfertigkeit gewaehlt.

            final SonderfertigkeitWert wert;
            wert = panelSonderfertigkeitWert.getGewaehltesElement();

            manager.einfuegenVorgegeben( wert );

            panelSonderfertigkeitWert = null;
        }

        if ( manager.hatSonderfertigkeitBonusOffen() ) {

            final SonderfertigkeitBonus bonus;
            bonus = manager.getSonderfertigkeitBonusOffen();

            final List< SonderfertigkeitWert > werte;
            werte = bonus.getWerte( manager.getHeld() );

            panelSonderfertigkeitWert = panelErschaffung.getPanelSonderfertigkeitWert();

            panelSonderfertigkeitWert.setElemente(
                    "Erschaffung.SonderfertigkeitWaehlen",
                    werte,
                    SonderfertigkeitWert.NAME_COMPARATOR );

            panelErschaffung.anzeigen( panelSonderfertigkeitWert );

            panelErschaffung.setWeiterAction( actionSonderfertigkeitBoniOffen );

        } else {

            // Alle Boni sind abgearbeitet. Es kann zum naechsten Schritt
            // ueber gegangen werden.
            schritt_3_2_VerbilligteSonderfertigkeitOffen();

        }
    }

    /**
     * Teilschritt 3.2 der Heldenerschaffung.
     * Hier werden die verbilligten Sonderfertigkeiten gewaehlt.
     */
    private void schritt_3_2_VerbilligteSonderfertigkeitOffen() {

        if ( null != panelSonderfertigkeit ) {
            // Es wurde gerade eine Sonderfertigkeit gewaehlt.

            final Sonderfertigkeit sonderfertigkeit;
            sonderfertigkeit = panelSonderfertigkeit.getGewaehltesElement();

            manager.einfuegenVerbilligt( sonderfertigkeit );

            panelSonderfertigkeit = null;
        }

        if ( manager.hatVerbilligteSonderfertigkeitOffen() ) {

            final VerbilligteSonderfertigkeit verbilligung;
            verbilligung = manager.getVerbilligteSonderfertigkeitOffen();

            final Held held = manager.getHeld();

            final List< Sonderfertigkeit > sonderfertigkeiten;
            sonderfertigkeiten = verbilligung.getSonderfertigkeiten( held );

            panelSonderfertigkeit = panelErschaffung.getPanelSonderfertigkeit();

            panelSonderfertigkeit.setElemente(
                    "Erschaffung.verbilligteSonderfertigkeitWaehlen",
                    sonderfertigkeiten,
                    Sonderfertigkeit.NAME_COMPARATOR );

            panelErschaffung.anzeigen( panelSonderfertigkeit );

            panelErschaffung.setWeiterAction( actionVerbilligteSonderfertigkeit );

        } else {

            schritt_4_VorNachteilBoniOffen();
        }
    }

    /**
     * Teilschritt 4 der Heldenerschaffung.
     * Der Benutzer kann die Vor-/Nachteile waehlen.
     */
    private void schritt_4_VorNachteilBoniOffen() {

        if ( null != panelVorNachteilWert ) {
            // Es wurde gerade ein Vor-/Nachteil gewaehlt.

            final VorNachteilWert wert;
            wert = panelVorNachteilWert.getGewaehltesElement();

            manager.einfuegenVorgegeben( wert );

            panelVorNachteilWert = null;
        }

        if ( manager.hatVorNachteilBonusOffen() ) {

            final VorNachteilBonus bonus;
            bonus = manager.getVorNachteilBonusOffen();

            final List< VorNachteilWert > werte;
            werte = bonus.getWerte();

            panelVorNachteilWert = panelErschaffung.getPanelVorNachteilWert();

            panelVorNachteilWert.setElemente(
                    "Erschaffung.VorNachteilWaehlen",
                    werte,
                    VorNachteilWert.NAME_COMPARATOR );

            panelErschaffung.anzeigen( panelVorNachteilWert );

            panelErschaffung.setWeiterAction( actionVorNachteilBoniOffen );

        } else {

            // Alle Boni sind abgearbeitet. Es kann zum naechsten Schritt
            // ueber gegangen werden.
            schritt_5_0_Zauber();

        }
    }

    /**
     * Teilschritt 5.0 der Erschaffung. Der Held erhaelt seine Zauber.
     */
    private void schritt_5_0_Zauber() {

        if ( manager.hatZusaetzlicheLeitzauber() ) {

            panelLeitzauber = panelErschaffung.getPanelLeitzauber();

            panelLeitzauber.setDaten(
                    manager.getHeld(),
                    manager.getAnzahlZusaetzlicheLeitzauber() );

            panelErschaffung.anzeigen( panelLeitzauber );

            panelErschaffung.setWeiterAction( actionLeitzauber );

        } else {

            schritt_6_Steigern();
        }
    }

    /**
     * Teilschritt 5.1 der Erschaffung, der Benutzer kann zusaetzliche
     * Leitzauber fuer den Helden waehlen.
     */
    private void schritt_5_1_LeitzauberGewaehlt() {

        for( final Zauber z: panelLeitzauber.getGewaehlteZauber() ) {

            manager.einfuegenLeitzauber( z );
        }

        schritt_6_Steigern();
    }


    /** Geht zum letzten Schritt der Erschaffung weiter. */
    private void schritt_6_Steigern() {

        panelErschaffung.setWeiterAction( actionBeenden );

        // muss hier zum 2ten mal ausgefuehrt werden, da sich die
        // Modifikationen auf die Eigenschaften durch die erhaltenen
        // Vor-/Nachteile geaendert haben koennen.
        manager.einstellenMindeststufe();

        final PanelSteigernErschaffung panelSteigern =
            panelErschaffung.getPanelSteigern();

        panelSteigern.setManager( manager );

        panelErschaffung.anzeigen( panelSteigern );

        hauptpanel.anpassenGroesse();
    }

    /** Startet eine neue Erschaffung. */
    public final void starten() {

        if ( ! hauptpanel.kannPanelWechseln() ) {
            
            return;
        }
        
        panelErschaffung = hauptpanel.getPanelErschaffung();
        panelLeitzauber = null;
        panelSonderfertigkeit = null;
        panelSonderfertigkeitWert = null;
        panelTalentBonus = null;
        panelVorNachteilWert = null;
        panelZweitsprache = null;

        manager = new ErschaffungManager();
        panelErschaffung.setManager( manager );
        
        final PanelGp panelGp = panelErschaffung.getPanelGp();

        panelGp.setManager( manager );

        hauptpanel.anzeigen( panelErschaffung );

        schritt_1_Bausteine();
    }

    /**
     * Eine Hilfsfunktion fuer die Entwicklung und Debugging. Erschafft einen
     * Helden mit den uebergebenen Bausteinen. Die Boni bei denen eine Auswahl
     * getroffen werden muss, werden ignoriert.
     */
    public final ErschaffungManager debugSchnellerschaffung(
            final String rasseId,
            final String kulturId,
            final String professionId ) {

        manager = new ErschaffungManager();

        final Held held = manager.getHeld();

        final Datenspeicher speicher = Datenspeicher.getInstanz();
        
        held.setRasse( speicher.getRasse( rasseId ) );
        held.setKultur( speicher.getKultur( kulturId ) );
        held.setProfession( speicher.getProfession( professionId ) );

        manager.anwendenSonderfertigkeiten();
        manager.anwendenTalentBoniLeittalente();
        manager.anwendenVorNachteile();
        manager.anwendenZauber();
        manager.einstellenMindeststufe();

        held.setWirdErschaffen( false );

        return manager;
    }

    /**
     * Beendet die Erschaffung nach dem Steigern.
     *
     * @author Bernhard Jung
     *
     */
    private class ActionBeenden extends AbstractAction implements Uebersetzbar {

        ActionBeenden() {
            super( I18n.i18n( "Erschaffung.Beenden.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Beenden.Mnemonic" ) ) );
        }

        public void actionPerformed( final ActionEvent e ) {

            final Held held = manager.getFertigenHeld();

            final PanelSteigern panelSteigern = hauptpanel.getPanelSteigern();

            panelSteigern.setHeld( held, true );

            hauptpanel.anzeigen( panelSteigern );
        }
        
        public void neuUebersetzen() {
            
            putValue( NAME, I18n.i18n( "Erschaffung.Beenden.Name" ) );
            
            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Beenden.Mnemonic" ) ) );
        }
    }

    /**
     * Die Action die verwendet wird wenn die zusaetzlichen Leitzauber gewaehlt
     * werden.
     */
    private class ActionLeitzauber 
    extends AbstractAction 
    implements Uebersetzbar {

        public ActionLeitzauber() {
            super( I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }

        public void actionPerformed( final ActionEvent e ) {
            schritt_5_1_LeitzauberGewaehlt();
        }

        public void neuUebersetzen() {
            
            putValue( NAME, I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }
    }

    /**
     * Die Action die verwendet wird wenn nacheinander die offenen Auswahlen
     * fuer die Sonderfertigkeiten abgearbeitet werden.
     */
    private class ActionSonderfertigkeitBoniOffen 
    extends AbstractAction
    implements Uebersetzbar {

        public ActionSonderfertigkeitBoniOffen() {
            super( I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }

        public void actionPerformed( final ActionEvent e ) {
            schritt_3_1_SonderfertigkeitBoniOffen();
        }

        public void neuUebersetzen() {
            
            putValue( NAME, I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }
    }

    /**
     * Die Action die verwendet wird wenn nacheinander die offenen Talent-Boni
     * abgearbeitet werden, die der Held durch Rasse und Kultur erhalten hat.
     */
    private class ActionTalentBoniOffen 
    extends AbstractAction
    implements Uebersetzbar {

        public ActionTalentBoniOffen() {
            super( I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }

        public void actionPerformed( final ActionEvent e ) {
            schritt_3_2_TalentBoniOffen();
        }

        public void neuUebersetzen() {
            
            putValue( NAME, I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }
    }

    /**
     * Die Action die verwendet wird wenn nacheinander die offenen Talent-Boni,
     * die der Held durch seine Profession erhalten hat, abgearbeitet werden.
     */
    private class ActionTalentBoniOffenProfession 
    extends AbstractAction 
    implements Uebersetzbar {

        public ActionTalentBoniOffenProfession() {
            super( I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }

        public void actionPerformed( final ActionEvent e ) {
            schritt_3_3_TalentBoniOffenProfession();
        }

        public void neuUebersetzen() {
            
            putValue( NAME, I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }
    }

    /**
     * Die Action die zum 2ten Schritt der Erschaffung, der Auswahl der
     * Talent-Boni, fuehrt.
     */
    private class ActionTalentBoni 
    extends AbstractAction 
    implements Observer, Uebersetzbar {

        private ActionTalentBoni() {
            super( I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );

            setEnabled( false );
        }

        public final void actionPerformed( final ActionEvent e ) {
            schritt_2_BoniAnwenden();
        }

        public void neuUebersetzen() {
            
            putValue( NAME, I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }

        /**
         * @param o der Held der ueberwacht wird.
         * @param arg ungenutzt.
         */
        public final void update( final Observable o, final Object arg ) {

            final Held held = (Held) o;

            setEnabled( ( held.getRasse()   != Rasse.LEERE_RASSE )
                  && ( held.getKultur()     != Kultur.LEERE_KULTUR )
                  && ( held.getProfession() != Profession.LEERE_PROFESSION ) );
        }
    }

    /**
     * Die Action die verwendet wird wenn die offenen Auswahlen
     * fuer die verbilligte Sonderfertigkeit abgearbeitet werden.
     */
    private class ActionVerbilligteSonderfertigkeit 
    extends AbstractAction
    implements Uebersetzbar {

        public ActionVerbilligteSonderfertigkeit() {
            super( I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }

        public void actionPerformed( final ActionEvent e ) {
            schritt_3_1_SonderfertigkeitBoniOffen();
        }

        public void neuUebersetzen() {
            
            putValue( NAME, I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }
    }

    /**
     * Die Action die verwendet wird um die offenen Auswahlen fuer die
     * Vor-/Nachteile abzuarbeiten.
     */
    private class ActionVorNachteilBoniOffen 
    extends AbstractAction
    implements Uebersetzbar {

        public ActionVorNachteilBoniOffen() {
            super( I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }

        public void actionPerformed( final ActionEvent e ) {
            schritt_4_VorNachteilBoniOffen();
        }

        public void neuUebersetzen() {
            
            putValue( NAME, I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }
    }

    /**
     * Die Action die verwendet wird um die Zweitsprache des Helden zu waehlen.
     */
    private class ActionZweitsprache 
    extends AbstractAction 
    implements Uebersetzbar {

        public ActionZweitsprache() {
            super( I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }

        public void actionPerformed( final ActionEvent e ) {
            schritt_3_1_Zweitsprache();
        }

        public void neuUebersetzen() {
            
            putValue( NAME, I18n.i18n( "Erschaffung.Weiter.Name" ) );

            putValue( 
                    MNEMONIC_KEY, 
                    Util.getKeyCode( I18n.i18n( "Erschaffung.Weiter.Mnemonic" ) ) );
        }
    }

}
