/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.i18n;

import java.util.Locale;
import java.util.ResourceBundle;

import de.bernhardjung.dsaprogramm.util.Einstellungen;

public class I18n {
    
    /** Das ResourceBundle in dem die Lokalisierten Texte gespeichert sind. */
    private static ResourceBundle RESOURCE_BUNDLE;
    
    static {
        
        final Einstellungen einstellungen = Einstellungen.getInstanz();
        
        final Locale sprache = einstellungen.getSprache();
        
        RESOURCE_BUNDLE = ResourceBundle.getBundle(
                "de.bernhardjung.dsaprogramm.i18n.Lokalisierung",
                sprache );
    }
    
    /**
     * Prueft ob es fuer einen Schluessel einen Lokalisierungstext gibt.
     * 
     * @param schluessel zu dem der Text gesucht ist.
     * @return {@code true}, wenn ein Text zum Schlussel existiert, sonst
     *      {@code false}.
     */
    public static boolean hatLokalisierungstext( 
            final String schluessel ) {
        
        return RESOURCE_BUNDLE.containsKey( schluessel );
    }
    
    /**
     * Liefert zu einem gegebenen Schluessel einen lokalisierten Text.
     * 
     * @param schluessel zu dem der lokalisierte Text gesucht ist.
     * @return den zum {@code schluessel} gehoerenden Text. Wenn kein zum 
     *      Schluessel gehoerender Text gefunden wird, wird {@code null} 
     *      geliefert.
     */
    public static String i18n( final String schluessel ) {

        if ( hatLokalisierungstext( schluessel ) ) {
            
            return RESOURCE_BUNDLE.getString( schluessel );
            
        } else {
        
            System.out.printf(
                    "Lokaliserungstext fuer Schluessel %1$s nicht gefunden." +
                            "\n",
                    schluessel );

        }

        return schluessel;
    }

    /**
     * Versucht einen lokalisierten Text anhand eines Schluessel zu laden.
     * Wenn kein Text gefunden wird, wird der Schluessel zurueckgegeben.
     * 
     * @param schluessel fuer den ein lokalisierter Text gesucht wird.
     * @return Der lokalisierte Text oder der Schluessel.
     */
    public static String i18nVersuch( final String schluessel ) {
        
        if ( hatLokalisierungstext( schluessel ) ) {
            
            return i18n( schluessel );
        
        } else {
            
            return schluessel;
        }
    }
    
    /**
     * Laedt die Resource-Datei mit den Uebersetzungen neu, z.B. wenn die
     * Sprache fuer das Programm geaendert wurde.
     */
    public static void neuLaden() {
        
        final Einstellungen einstellungen = Einstellungen.getInstanz();
        
        RESOURCE_BUNDLE = ResourceBundle.getBundle(
                "de.bernhardjung.dsaprogramm.i18n.Lokalisierung",
                einstellungen.getSprache() );
    }
}
