/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.talente;

import java.util.ArrayList;
import java.util.List;

import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.basis.Lernmethode;
import de.bernhardjung.dsaprogramm.util.Util;

public class TalentWert implements Cloneable {

    /**
     * Gibt an ob das Talent aktiviert ist.
     * Das wird gebraucht fuer Spezial-Talente mit negativer Stufe.
     * Ein solches Talent muss erst mal auf Stufe 0 gesteigert werden.
     * Auf Stufe 0 gilt es dann aber noch nicht als aktiviert. Das muss noch
     * extra gemacht und mit AP bezahlt werden.
     * Das wird dann durch setzen dieser Variable gemacht.
     */
    private boolean aktiviert;

    /** Die AP die zum Aktivieren des Talents aufgewendet wurden. */
    private int aktivierungskosten;

    /**
     * Speichert bei einer Kampftechnik wie viele Punkte der Stufe auf die
     * Attacke verteilt werden.
     */
    private int attacke;

    /** Der Held der den Wert besitzt. */
    private Held held;

    /**
     * Speichert die AP, die aufgewendet wurden, um den Talentwert zu steigern.
     * An Position i stehen die Kosten, die fuer den i-ten Punkt, in der
     * unmodifizierten Stufe aufgewendet wurden.
     */
    private List< Integer > kosten;

    /**
     * Speichert die AP, die aufgewendet wurden, um den Talentwert, im negativen
     * Bereich, zu steigern. An Position i stehen die Punkte die fuer den
     * |i|-ten Punkt in der unmodifizierten Stufe aufgewendet wurden.
     */
    private List< Integer > kostenNegativ;

    /**
     * Gibt an ob der Held eine spezielle Erfahrung auf den Wert erhalten hat.
     */
    private boolean spezielleErfahrung;

    /** Das Talent des Wertes. */
    private final Talent talent;

    /**
     * Die Unmodifizierte Stufe des Talentwerts.
     * Modifikationen durch Vor-/Nachteile muessen hier noch eingerechnet
     * werden.
     */
    private int unmodifizierteStufe;

    /**
     * Erzeugt einen neuen, nicht vorgegebenen Talentwert mit einer Stufe von
     * 0.
     *
     * @param dasTalent des Wertes. Das Argument darf nicht {@code null} sein.
     */
    public TalentWert( final Talent dasTalent ) {

        this( dasTalent, 0 );
    }

    /**
     * Erzeugt einen neuen, nicht vorgegebenen Talentwert.
     *
     * @param dasTalent des Wertes. Das Argument darf nicht {@code null} sein.
     * @param dieStufe Die unmodifizierte Stufe des Wertes.
     */
    public TalentWert( final Talent dasTalent, final int dieStufe ) {
        this( dasTalent, dieStufe, 0 );
    }


    /**
     * Erzeugt einen neuen, nicht vorgegebenen Talentwert.
     *
     * @param dasTalent
     *            des Wertes. Das Argument darf nicht {@code null} sein.
     * @param dieStufe
     *            Die unmodifizierte Stufe des Wertes.
     * @param dieAktivierungskosten
     *            die zum aktivieren des Talents bezahlt wurden.
     */
    public TalentWert(
            final Talent dasTalent,
            final int dieStufe,
            final int dieAktivierungskosten ) {

        if ( null == dasTalent ) {
            throw new IllegalArgumentException( "null ist als Argument fuer " +
                    "dasTalent nicht erlaubt." );
        }

        // Ein Spezial-Talent mit einer Stufe unter 0 gilt erst mal als nicht
        // aktiviert.
        aktiviert = ( 0 <= dieStufe );

        aktivierungskosten = dieAktivierungskosten;

        attacke = 0;

        kosten = new ArrayList< Integer >();

        kostenNegativ = new ArrayList< Integer >( 3 );

        talent = dasTalent;

        unmodifizierteStufe = dieStufe;

        for ( int i = 0; i <= unmodifizierteStufe ; ++i ) {

            kosten.add( 0 );
        }

        kostenNegativ.add( 0 );

        if ( unmodifizierteStufe < 0 ) {

            for ( int i = 1, n = Math.abs( unmodifizierteStufe )
                    ; i <= n
                    ; ++i ) {

                kostenNegativ.add( 0 );
            }
        }
    }

    /**
     * Hebt/Senkt die Stufe des Talentwerts.
     *
     * @param bonus der zur Stufe addiert wird.
     */
    public final void addiereStufe( final int bonus ) {

        if ( 0 < bonus ) {

            for ( int i = 0 ; i < bonus ; ++i ) {

                ++unmodifizierteStufe;

                setKosten( 0 );
            }

        } else if ( 0 > bonus ) {

            for ( int i = 0, n = Math.abs( bonus ) ; i < n ; ++i ) {

                entferneKosten();

                --unmodifizierteStufe;
            }
        }

        if ( null != held ) {

            aktualisiereAktiviert();

            held.notifyObservers( talent );
        }
    }

    /**
     * Aktualisiert den Wert von {@link #aktiviert} entsprechend der Stufe
     * des Talents.
     */
    private boolean aktualisiereAktiviert() {

        final int stufe;

        if ( null != held ) {

            stufe = getStufe( held );

        } else {

            stufe = unmodifizierteStufe;

        }

        if ( 0 > stufe ) {
            // Wenn die Stufe unter 0 sinkt soll das Talent erst mal als nicht
            // aktiviert gelten.
            aktiviert = false;

        } else if ( 0 < stufe ) {
            // Bei positiver Stufe ist das Talent automatisch aktiviert.
            aktiviert = true;

        }

        return aktiviert;
    }

    /**
     * @return eine Kopie des Wertes.
     */
    @Override
    public final TalentWert clone() {

         try {

             final TalentWert kopie = (TalentWert) super.clone();

             kopie.kosten = new ArrayList< Integer >( kosten );

             kopie.kostenNegativ = new ArrayList< Integer >( kostenNegativ );

             return kopie;

         } catch (final CloneNotSupportedException e) {

         }

        return null;
    }
    
    @Override
    public boolean equals( Object obj ) {
        
        if ( null == obj 
               || ! obj.getClass().equals( getClass() ) ) {
            
            return false;
        }
        
        final TalentWert wert = (TalentWert) obj;
        
        return Util.equals( 
                aktiviert, wert.aktiviert,
                aktivierungskosten, wert.aktivierungskosten,
                attacke, wert.attacke,
                kosten, wert.kosten,
                kostenNegativ, wert.kostenNegativ,
                spezielleErfahrung, wert.spezielleErfahrung,
                talent, wert.talent,
                unmodifizierteStufe, wert.unmodifizierteStufe );
    }

    private int entferneKosten() {

        if ( 0 <= unmodifizierteStufe ) {

            return kosten.remove( unmodifizierteStufe );

        } else {

            final int index = Math.abs( unmodifizierteStufe );

            final int ap;

            if ( index < kostenNegativ.size() ) {

                ap = kostenNegativ.get( index );

                kostenNegativ.set( index, 0 );

            } else {

                ap = 0;

                kostenNegativ.add( ap );

            }

            return ap;
        }
    }

    /**
     * @return die AP, die zum Aktivieren des Talents bezahlt wurden.
     */
    public final int getAktivierungskosten() {
        return aktivierungskosten;
    }

    public final int getApKosten() {

        int apKosten = aktivierungskosten;

        for ( final int ap : kosten ) {

            apKosten += ap;
        }

        for ( final int ap : kostenNegativ ) {

            apKosten += ap;
        }

        return apKosten;
    }

    /**
     * @return wie viele Punkte der Stufe auf Attacke verteilt wurden.
     */
    public final int getAttacke() {
        return attacke;
    }

    /**
     * @param held dessen Parade-Wert gefragt ist.
     * @return die Anzahl Punkte der Stufe die auf Parade verteilt wurden.
     */
    public final int getParade( final Held held ) {
        return getStufe( held ) - attacke;
    }

    /* *
     * @param held
     *            der das Talent steigern will.
     * @return die AP die bezahlt werden muessen um das Talent um einen Punkt zu
     *         Steigern.
     * /
    public int getSteigerungskosten( final Held held ) {
        return talent.getSteigerungskosten( held, getStufe( held ) );
    }
*/
    /**
     * @param held
     *            der das Talent steigern will.
     * @param lernmethode
     *            mit der das Talent gesteigert werden soll.
     * @return die AP die zu bezahlen sind um das Talent um einen Punkt
     *         anzuheben.
     */
    public final int getSteigerungskosten(
            final Held held,
            Lernmethode lernmethode ) {

        if ( spezielleErfahrung ) {

            lernmethode = Lernmethode.SPEZIELLE_ERFAHRUNG;
        }

        return talent.getSteigerungskosten( held, getStufe(held), lernmethode );
    }

    /**
     * Liefert die Stufe des Helden in dem Talent, wobei alle Modifikationen
     * durch Vor-/Nachteile eingerechnet werden.
     *
     * @param held der das Talent besitzt und dessen Stufe in dem Talent
     *      gefragt ist.
     * @return die Stufe des Helden in dem Talent.
     * @throws UnsupportedOperationException wenn nicht mittels
     *      {@link #setHeld(Held)} ein Held gesetzt wurde.
     */
    public final int getStufe( final Held held ) {
        return unmodifizierteStufe + talent.getModifikation( held );
    }

    /**
     * @return das Talent des Wertes.
     */
    public final Talent getTalent() {
        return talent;
    }

    /**
     * @see #getStufe(Held)
     * @return die unmodifiziert Stufe des Talents. Modifikationen durch
     *      Vor-/Nachteile sind hier noch nicht eingerechnet.
     */
    public final int getUnmodifizierteStufe() {
        return unmodifizierteStufe;
    }

    /**
     * @return {@code true}, wenn der Held eine Spezielle Erfahrung auf das
     *         Talent erhalten hat, sonst {@code false}.
     */
    public final boolean hatSpezielleErfahrung() {
        return spezielleErfahrung;
    }

    /**
     * @return {@code true} wenn das Talent als aktivert gilt, sonst
     *      {@code false}.
     */
    public final boolean istAktiviert() {

        return talent.istBasisTalent() // Basis-Talente sind immer aktivert.
            || aktiviert;
    }

    /**
     * @param istAktiviert
     *            {@code true} wenn das Talent aktiviert ist, sonst
     *            {@code false}.
     */
    public final void setAktiviert( final boolean istAktiviert ) {

        if ( istAktiviert != aktiviert ) {

            aktiviert = istAktiviert;

            if ( null != held ) {
                held.notifyObservers( talent );
            }
        }
    }

    /**
     * @param apKosten die Aufgewendet wurden um das Talent zu aktivieren.
     */
    public final void setAktivierungskosten( final int apKosten ) {
        aktivierungskosten = apKosten;
    }

    /**
     * @param punkte der Talentstufe die auf Attacke verteilt werden.
     */
    public final void setAttacke( final int punkte ) {

        if ( punkte != attacke ) {

            attacke = punkte;

            if ( null != held ) {
                held.notifyObservers( talent );
            }
        }
    }

    /**
     * Bestimmt den Helden der den Wert besitzt. Der Held wird im folgenden
     * ueber Aenderungen an diesem Wert informiert. Dazu wird die Methode
     * {@link Held#notifyObservers()} aufgerufen.
     *
     * @param derHeld der den Wert bestizt.
     */
    public final void setHeld( final Held derHeld ) {
        held = derHeld;
    }

    private void setKosten( final int apKosten ) {

        if ( 0 <= unmodifizierteStufe ) {

            kosten.add( apKosten );

        } else {

            final int index = Math.abs( unmodifizierteStufe );

            if ( kostenNegativ.size() <= index ) {

                kostenNegativ.add( apKosten );

            } else {

                kostenNegativ.set( index, apKosten );

            }
        }
    }

    /**
     * @param erfahrung {@code true}, wenn der Held eine spezielle Erfahrung auf
     *            das Talent erhalten hat, sonst {@code false}.
     */
    public final void setSpezielleErfahrung( final boolean erfahrung ) {

        if ( erfahrung != spezielleErfahrung ) {

            spezielleErfahrung = erfahrung;

            if ( null != held ) {
                held.notifyObservers( talent );
            }
        }
    }

    /**
     * Setzt die Stufe des Talents auf den angegebenen Wert.
     *
     * @param stufe die das Talent haben soll.
     */
    public final void setStufe( final int stufe ) {

        int modifikation = 0;
        if ( null != held ) {
            modifikation = talent.getModifikation( held );
        }

        final int neueUnmodifizierteStufe = stufe - modifikation;

        // Nur wenn der Wert auch wirklich anders ist.
        if ( neueUnmodifizierteStufe != unmodifizierteStufe ) {

            if ( neueUnmodifizierteStufe > unmodifizierteStufe ) {

                while ( neueUnmodifizierteStufe > unmodifizierteStufe ) {

                    ++unmodifizierteStufe;

                    setKosten( 0 );
                }

            } else if ( neueUnmodifizierteStufe < unmodifizierteStufe ) {

                while ( neueUnmodifizierteStufe < unmodifizierteStufe ) {

                    entferneKosten();

                    --unmodifizierteStufe;
                }
            }

            aktualisiereAktiviert();

            if ( null != held ) {
                held.notifyObservers( talent );
            }
        }
    }

    /**
     * Setzt die Stufe des Talents auf den uebergebenen Wert.
     *
     * @param held
     *            der das Talent steigern will.
     * @param stufe
     *            die das Talent erhalten soll.
     * @param lernmethode
     *            die beim Steigern verwendet wird.
     * @return die AP-Kosten die aufgebracht werden muessen (positiver Wert), um
     *         die Stufe des Talents auf die uebergebenen Stufe zu steigern,
     *         oder die AP die frei werden (negativer Wert), wenn das Talent
     *         gesenkt wird.
     */
    public final int setStufe(
            final Held held,
            final int stufe,
            final Lernmethode lernmethode ) {

        // die gesmaten AP die zum Steigern benoetigt werden, bzw. die AP die
        // durch das Senken der Stufe frei werden.
        int summe = 0;

        final int modifikation = talent.getModifikation( held );

        final int neueUnmodifizierteStufe = stufe - modifikation;

        final int ausgangsstufe = getStufe( held );

        if ( neueUnmodifizierteStufe > unmodifizierteStufe ) {

            int vorgegUnmodStufe = 0;

            if ( held.hatVorgegeben( talent ) ) {

                vorgegUnmodStufe =
                    held.getVorgegeben( talent ).getUnmodifizierteStufe();
            }

            while ( ( unmodifizierteStufe < vorgegUnmodStufe )
                 && ( unmodifizierteStufe < neueUnmodifizierteStufe ) ) {

                ++unmodifizierteStufe;
                // Steigern bis zur vorgegebenen Stufe ist kostenlos.
                setKosten( 0 );
            }


            while ( unmodifizierteStufe < neueUnmodifizierteStufe ) {

                final int apKosten;

                if ( spezielleErfahrung ) {

                    apKosten = talent.getSteigerungskosten(
                            held,
                            getStufe( held ),
                            Lernmethode.SPEZIELLE_ERFAHRUNG );
                } else {

                    apKosten = talent.getSteigerungskosten(
                            held,
                            getStufe( held ),
                            lernmethode );
                }

                spezielleErfahrung = false;

                ++unmodifizierteStufe;

                setKosten( apKosten );

                summe += apKosten;

            }

            // Wenn das Talent nicht vorgegeben ist:
            if ( ( ! held.hatVorgegeben( talent ) )
              // Die Stufe wurde aus dem nicht positiven Bereich
              && ( 0 >= ausgangsstufe )
              // in den postiven Bereich angehoben und
              && ( 0 < stufe )
              // und das Talent war zuvor nicht aktiviert.
              && ( ! istAktiviert() ) ) {
                // Daher muss es jetzt noch aktiviert werden und dafuer sind die
                // Aktivierungskosten zu zahlen.
                final int aktivierungskosten =
                    talent.getAktivierungskosten( held, lernmethode );

                setAktivierungskosten( aktivierungskosten );

                summe += aktivierungskosten;
            }

            aktualisiereAktiviert();

        } else if ( neueUnmodifizierteStufe < unmodifizierteStufe ) {

            while ( unmodifizierteStufe > neueUnmodifizierteStufe ) {

                summe -= entferneKosten();

                --unmodifizierteStufe;
            }

            aktualisiereAktiviert();

            if ( ! istAktiviert() ) {
                // Das Talent ist aktuell nicht aktiviert.
                // Wenn es zuvor aktiviert war, so muessen die
                // Aktivierungskosten gutgeschrieben werden.
                summe -= getAktivierungskosten();

                setAktivierungskosten( 0 );
            }
        }

        return summe;
    }
    
    /** 
     * @return Eine Ausgabe des Talentwerts fuers Debugging.
     */
    @Override
    public String toString() {
        
        return String.format( "%s %d", talent.getName(), unmodifizierteStufe );
    }
}
