/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.talente;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.gui.DialogFehlermeldung;

public class TalentWerte implements Cloneable {

    /** Der Held dem die Talentwerte gehoeren. */
    private final Held held;

	/** Alle gespeicherten Talente. */
	private Map< Talent, TalentWert > talente;

    /**
     * Erzeugt eine leere Liste fuer Talente
     */
    public TalentWerte() {
        this( null );
    }

    /**
     * Erzeugt eine leere Liste fuer Talente.
     *
     * @param derHeld dessen Talente in dieser Liste gespeichert werden.
     *       Der Held wird ueber Aenderungen an der Liste und an den Werten
     *       informiert. Dazu wird bei jeder Veraenderung die Methode
     *       {@link Held#notifyObservers()} aufgerufen.
     *       Darf {@code null} sein.
     */
	public TalentWerte( final Held derHeld ) {

        held = derHeld;

		talente = new HashMap< Talent, TalentWert >();
	}

	/**
	 * @return Eine Deep Copy der Collection.
	 */
	@Override
	public TalentWerte clone() {
	    
	    TalentWerte clone = null;
	    
	    try {
            
	        clone = (TalentWerte) super.clone();
            
	        clone.talente = cloneTalente();
	        
        } catch ( CloneNotSupportedException e ) {
            
            DialogFehlermeldung.zeigeDialog( e );
        }
        
        return clone;
	}
	
	private Map< Talent, TalentWert > cloneTalente() {
	    
	    final Map< Talent, TalentWert > clone = 
	        new HashMap< Talent, TalentWert >( talente.size() );
	    
	    for ( final Talent talent : talente.keySet() ) {
	        
	        clone.put( 
	                talent, 
	                talente.get( talent ).clone() );
	    }
	    
	    return clone;
	}
	
	/**
	 * Hebt den Talenwert eines Talents.
     * Wenn das Talent in der Liste noch nicht existiert wird es eingefuegt,
     * wenn das Talent schon in der Liste vorhanden ist wird seine Stufe um die
     * unmodifizierte Stufe von {@code wert} erhoeht.
	 *
	 * @param wert Der Bonus den das Talent erhalten soll.
     * @throws NullPointerException wenn {@code null} uebergeben wird.
	 */
	public void einfuegen( final TalentWert wert ) {

		final Talent talent = wert.getTalent();

		if ( ! enthaelt( talent ) ) {

            final TalentWert neuerWert = wert.clone();

            neuerWert.setHeld( held );

			talente.put( talent, neuerWert );

			if ( null != held ) {

			    held.notifyObservers( talent );
			}

		} else {

		    final TalentWert enthaltenerWert = getWert( talent );

		    enthaltenerWert.addiereStufe( wert.getUnmodifizierteStufe() );

		    final int at = enthaltenerWert.getAttacke();

		    enthaltenerWert.setAttacke( at + wert.getAttacke() );
		}
	}

	/**
	 * Entfernt ein Talent aus der Liste, so vorhanden.
	 *
	 * @param talent das entfernt werden soll.
	 */
	public void entfernen( final Talent talent ) {

        final TalentWert wert = talente.remove( talent );

        if ( null != wert
                && null != held ) {

            wert.setHeld( null );

            held.notifyObservers( talent );
        }
	}

	/**
	 * Prueft ob ein Talent in der Liste gespeichert ist.
	 *
	 * @param talent das gesucht wird.
	 * @return {@code true}, wenn das Talent gespeichert ist, sonst
	 * 			{@code false}.
	 */
	public boolean enthaelt( final Talent talent ) {

		return talente.containsKey( talent );
	}

	@Override
	public boolean equals( Object obj ) {
	    
	    if ( null == obj 
	            || ! obj.getClass().equals( getClass() ) ) {
	        
	        return false;
	    }
	     
	    final TalentWerte werte = (TalentWerte) obj;
	    
	    return talente.equals( werte );
	}
	
	/**
     * Liefert eine Liste aller Talente zu einer bestimmten Talentgruppe, die in
     * dieser Talentliste gespeichert sind.
     *
     * @param talentgruppe
     *            deren Talente gesucht sind.
     * @return eine Liste mit allen gespeicherten Talenten zur gesuchten
     *         Talentgruppe. Wenn dieses Objekt einem Helden zugeordnet ist, so
     *         werden nur die fuer den Held relevanten Talente geliefert, d.h.
     *         Spezial-Talente in denen der Held eine negative Stufe hatte, die
     *         er aber auf Stufe 0 gesteigert hat, ohne sie zu aktivieren,
     *         werden von der Methode nicht geliefert, auch wenn sie in diesem
     *         Objekt gespeichert sind. Spezial-Talente mit negativer Stufe
     *         werden sehr wohl geliefert.
     */
	public List< Talent > get( final Talentgruppe talentgruppe ) {

		final List< Talent > liste = new ArrayList< Talent >();

		for ( final Map.Entry< Talent, TalentWert > entry : talente.entrySet()){

		    final Talent talent = entry.getKey();

		    final TalentWert wert = entry.getValue();

			if ( talent.getTalentgruppe().equals( talentgruppe )
			  && ( ( held == null )
			    || ( wert.getStufe( held ) != 0 )
			    || ( wert.istAktiviert() ) ) ) {

				liste.add( talent );
			}
		}

		return liste;
	}

	/**
	 * @return die Anzahl der gespeicherten Talente.
	 */
	public int getAnzahlTalente() {
		return talente.size();
	}

	/**
     * @return eine Liste aller gespeicherten Talente. Wenn dieses Objekt einem
     *         Helden zugeordnet ist, so werden nur die fuer den Held relevanten
     *         Talente geliefert, d.h. Spezial-Talente in denen der Held eine
     *         negative Stufe hatte, die er aber auf Stufe 0 gesteigert hat,
     *         ohne sie zu aktivieren, werden von der Methode nicht geliefert,
     *         auch wenn sie in diesem Objekt gespeichert sind. Spezial-Talente
     *         mit negativer Stufe werden sehr wohl geliefert.
     */
	public List< Talent > getTalente() {

	    final List< Talent > liste;

	    if ( held == null ) {

	        liste = new ArrayList< Talent >( talente.keySet() );

	    } else {

	        liste = new ArrayList< Talent >();

	        for ( final Map.Entry< Talent, TalentWert > entry
	                : talente.entrySet() ) {

	            final TalentWert wert = entry.getValue();

	            final int stufe = wert.getStufe( held );

	            if ( ( stufe != 0 ) || wert.istAktiviert() ) {

	                liste.add( entry.getKey() );
	            }
	        }

	    }

	    return liste;
	}

	/**
	 * @return den zum Talent gehoerenden Talentwert oder {@code null},
	 * 		wenn kein passender TalentWert vorhanden ist. Es wird das
     *      Objekt geliefert, das in dieser Liste gespeichert ist. Aenderungen
     *      an diesem Objekt wirken sich also direkt aus.
	 */
	public TalentWert getWert( final Talent talent ) {
		return talente.get( talent );
	}

}
