/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.vornachteile.voraussetzungen;

import java.util.Collection;
import java.util.Collections;

import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.basis.KonfliktBeschreibung;
import de.bernhardjung.dsaprogramm.basis.varianten.Variante;
import de.bernhardjung.dsaprogramm.i18n.I18n;
import de.bernhardjung.dsaprogramm.vornachteile.VorNachteil;
import de.bernhardjung.dsaprogramm.vornachteile.VorNachteilWert;

class VoraussetzungStandard extends Voraussetzung {

    /**
     * @return eine leere Liste.
     */
    @Override
    @SuppressWarnings( "unchecked" )
    public Collection<Object> getAbhaengigkeiten() {
        return Collections.EMPTY_LIST;
    }

    /**
     * Prueft ob ein Held keine staerkere Auspraegung eines Vor-/Nachteils
     * besitzt.
     *
     * @param held
     *            der die staerkere Auspraegung nicht besitzen soll.
     * @param wert
     *            dessen staerkere Auspraegung(en) der Held nicht besitzen soll.
     * @return {@code true}, wenn der Held keine staerkere Auspraegung des
     *         Vor-/Nachteils besitzt, sonst {@code false}.
     */
    private boolean besitztKeineStaerkereAuspraegung(
            final Held held,
            final VorNachteilWert wert ) {

        VorNachteil vorNachteil = wert.getVorNachteil();

        while ( vorNachteil.hatStaerkereAuspraegung() ) {

            vorNachteil = vorNachteil.getStaerkereAuspraegung();

            final VorNachteilWert staerkereAuspraegung;

            if ( vorNachteil.hatVarianten() ) {

                final Variante variante = wert.getVariante();

                staerkereAuspraegung =
                    new VorNachteilWert( vorNachteil, variante );
                
            } else {

                staerkereAuspraegung = new VorNachteilWert( vorNachteil );

            }

            if ( held.hat( staerkereAuspraegung ) ) {

                return false;
            }
        }

        return true;
    }

    /** {@inheritDoc} */
    @Override
    public final KonfliktBeschreibung pruefen(
            final Held held,
            final VorNachteilWert wert ) {

        final KonfliktBeschreibung beschreibung = new KonfliktBeschreibung();

        if ( held.hatVorgegeben( wert )
                && ( ! held.hat( wert ) )
                && ( ! held.istAbgebaut( wert ) ) ) {
            // Der Held sollte den Vor-/Nachteil besitzen, besitzt ihn aber
            // nicht.

            final VorNachteil vorNachteil = wert.getVorNachteil();

            if ( vorNachteil.hatStaerkereAuspraegung() ) {

                if ( besitztKeineStaerkereAuspraegung( held, wert ) ) {
                    // Der Held besitzt auch keine staerkere Auspraegung davon.
                    beschreibung.einfuegen( I18n.i18n( 
                            "VoraussetzungStandard.VorNachteil." +
                            "VorgegebenStaerkereAuspraegung" ) );
                }

            } else {

                beschreibung.einfuegen( I18n.i18n( 
                        "VoraussetzungStandard.VorNachteil.Vorgegeben" ) );

            }
        }

        return beschreibung;
    }

}
